# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.libs.pari (for factorization)
r"""
Cubic Hecke Base Rings

This module contains special classes of polynomial rings
(:class:`CubicHeckeRingOfDefinition` and :class:`CubicHeckeExtensionRing`)
used in the context of
:class:`cubic Hecke algebras
<sage.algebras.hecke_algebras.cubic_hecke_algebra.CubicHeckeAlgebra>`.

AUTHORS:

- Sebastian Oehms May 2020: initial version
"""
# ###########################################################################
#       Copyright (C) 2020 Sebastian Oehms <seb.oehms@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ###########################################################################
from sage.structure.category_object import normalize_names
from sage.structure.element import get_coercion_model
from sage.categories.action import Action
from sage.misc.verbose import verbose
from sage.misc.functional import cyclotomic_polynomial
from sage.misc.cachefunc import cached_method
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing_mpair
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.localization import Localization
from sage.algebras.splitting_algebra import solve_with_extension, SplittingAlgebra


# -----------------------------------------------------------------------------
# local helper functions
# -----------------------------------------------------------------------------
def normalize_names_markov(names, markov_trace_version):
    r"""
    Return a tuple of strings of variable names of length 3 resp. 4 (if
    ``markov_trace_version`` is ``True``) according to the given input names.

    INPUT:

    - ``names`` -- passed to :func:`~sage.structure.category_object.normalize_names`
    - ``markov_trace_version`` -- boolean; if set to ``True`` four names are
      expected the last of which corresponds to the writhe factor of the
      Markov trace

    EXAMPLES::

        sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
        sage: chbr.normalize_names_markov('a, b, c', False)
        ('a', 'b', 'c')
        sage: chbr.normalize_names_markov(('u', 'v', 'w', 's'), False)
        ('u', 'v', 'w')
    """
    if markov_trace_version:
        names = normalize_names(4, names)
    else:
        if isinstance(names, tuple):
            names = list(names)
        if isinstance(names, list) and len(names) > 3:
            names = normalize_names(3, names[0:3])
        else:
            names = normalize_names(3, names)
    return names


def register_ring_hom(ring_hom):
    r"""
    Register the given ring homomorphism as conversion map.

    EXAMPLES::

        sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
        sage: BR = chbr.CubicHeckeRingOfDefinition()
        sage: BR.create_specialization([E(5), E(7), E(3)])  # indirect doctest
        Universal Cyclotomic Field
        sage: _.convert_map_from(BR)
        Ring morphism:
          From: Multivariate Polynomial Ring in u, v, w
                  over Integer Ring localized at (w,)
          To:   Universal Cyclotomic Field
          Defn: u |--> E(5)
                v |--> E(7)
                w |--> E(3)
    """
    domain = ring_hom.domain()
    codomain = ring_hom.codomain()
    conversion_cached = codomain._is_conversion_cached(domain)

    if conversion_cached:
        test_map = codomain.convert_map_from(domain)
        try:
            if test_map != ring_hom:
                verbose('\nConversion:\n%s\n already exists and is different from:\n%s\n' % (test_map, ring_hom))
        except TypeError:
            verbose('\n Conversion:\n%s\n already exists and is not comparable to:\n%s\n' % (test_map, ring_hom))
    else:
        try:
            codomain.register_conversion(ring_hom)
        except ValueError:
            verbose('\nthe map:\n%s\ncannot be registered as conversion\n' % ring_hom)


# -----------------------------------------------------------------------------
# class for the Galois Group action on the generic extension ring corresponding
# to the cubic equation
# -----------------------------------------------------------------------------
class GaloisGroupAction(Action):
    r"""
    Action on a multivariate polynomial ring by permuting the generators.

    EXAMPLES::

        sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
        sage: from operator import mul
        sage: R.<x, y, z> = ZZ[]
        sage: G = SymmetricGroup(3)
        sage: p = 5*x*y + 3*z**2
        sage: R._unset_coercions_used()
        sage: R.register_action(chbr.GaloisGroupAction(G, R, op=mul))
        sage: s = G([2,3,1])
        sage: s*p
        3*x^2 + 5*y*z
    """
    def _act_(self, perm, pol):
        r"""
        Application of the action.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: from operator import mul
            sage: R.<x, y> = QQ[]
            sage: G = SymmetricGroup(2)
            sage: A = chbr.GaloisGroupAction(G, R, op=mul)
            sage: p = ~5*x*y**2 + 3*x**2
            sage: s = G([2,1])
            sage: A._act_(s, p)
            1/5*x^2*y + 3*y^2
        """
        if not self.is_left():
            perm, pol = pol, perm
        pol_dict = {}
        for key, value in pol.monomial_coefficients().items():
            newkey = [0] * len(key)
            for pos, k in enumerate(key):
                newkey[perm(pos + 1) - 1] = k
            pol_dict[tuple(newkey)] = value
        return self.domain()(pol_dict)


###############################################################################
# EXTENSION RING
# -----------------------------------------------------------------------------
# Definition of the generic extension ring for the cubic Hecke algebra as
# Laurent polynomial ring in 3 indeterminates over the cyclotomic field of a
# third root of unity This is the most general ring over which the cubic Hecke
# algebra is semi-simple. In opposite to the generic base ring class, this class
# does not inherits from UniqueRepresentation since _test_pickling fails
# -----------------------------------------------------------------------------
class CubicHeckeExtensionRing(LaurentPolynomialRing_mpair):
    r"""
    The generic splitting algebra for the irreducible representations of
    the cubic Hecke algebra.

    This ring must contain three invertible indeterminates (representing
    the roots of the cubic equation) together with a third root of unity
    (needed for the 18-dimensional irreducibles of the cubic Hecke algebra
    on 4 strands).

    Therefore this ring is constructed as a multivariate Laurent polynomial
    ring in three indeterminates over a polynomial quotient ring over the
    integers with respect to the minimal polynomial of a third root of unity.

    The polynomial quotient ring is constructed as instance of
    :class:`SplittingAlgebra`.

    INPUT:

    - ``names`` -- (default: ``'u,v,w'``) string containing the names of the
      indeterminates separated by ``,`` or a triple of strings each of which
      are the names of one of the three indeterminates
    - ``order`` -- string (default: ``'degrevlex'``); the term order; see also
      :class:`~sage.rings.polynomial.laurent_polynomial_ring.LaurentPolynomialRing_mpair`
    - ``ring_of_definition`` -- (optional) a :class:`CubicHeckeRingOfDefinition`
      to specify the generic cubic Hecke base ring over which ``self`` may be
      realized as splitting ring via the ``as_splitting_algebra`` method
    - ``third_unity_root_name`` -- string (default: ``'e3'``); for setting the
      name of the third root if unity of ``self``
    - ``markov_trace_version`` -- boolean (default: ``False``); if this is
      set to ``True`` then ``self`` contains one invertible indeterminate in
      addition which is meant to represent the writhe factor of a Markov trace
      on the cubic Hecke algebra and which default name is ``s``

    EXAMPLES::

        sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
        sage: chbr.CubicHeckeExtensionRing('a, b, c')
        Multivariate Laurent Polynomial Ring in a, b, c
          over Splitting Algebra of x^2 + x + 1
            with roots [e3, -e3 - 1]
          over Integer Ring
        sage: _.an_element()
        b^2*c^-1 + e3*a
    """
    def __init__(self, names, order='degrevlex', ring_of_definition=None, third_unity_root_name='e3', markov_trace_version=False):
        r"""
        Initialize ``self``.

        TESTS::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: TestSuite(ER).run()
        """
        # ----------------------------------------------------------------------
        # Setting connection with generic base ring (if given)
        # ----------------------------------------------------------------------
        self._ring_of_definition = None
        self._splitting_algebra = None

        if ring_of_definition is not None:
            if not isinstance(ring_of_definition, CubicHeckeRingOfDefinition):
                raise TypeError("generic base ring must be an instance of CubicHeckeRingOfDefinition")
            self._ring_of_definition = ring_of_definition

        # ----------------------------------------------------------------------
        # defining the base ring
        # note that we can't use ZZ.extension since it isn't possible to define
        # homomorphisms from orders in number fields, yet
        # ----------------------------------------------------------------------
        base_ring = SplittingAlgebra(cyclotomic_polynomial(3), [third_unity_root_name])

        # ----------------------------------------------------------------------
        # defining the ring itself
        # ----------------------------------------------------------------------
        self._names = normalize_names_markov(names, markov_trace_version)
        self._order = order

        pol_ring = PolynomialRing(base_ring, names=self._names, order=self._order, implementation=None)
        LaurentPolynomialRing_mpair.__init__(self, pol_ring)

        # ----------------------------------------------------------------------
        # setting Galois group action
        # ----------------------------------------------------------------------
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        from operator import mul
        self._galois_group = SymmetricGroup(3)
        galois_group_action = GaloisGroupAction(self._galois_group, self, op=mul)
        self._unset_coercions_used()
        self.register_action(galois_group_action)

        # ----------------------------------------------------------------------
        # Init of data used on demand
        # ----------------------------------------------------------------------
        self._mirror = None

    ############################################################################
    # overloaded inherited methods
    ############################################################################
    def construction(self):
        r"""
        Return ``None`` since this construction is not functorial.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: ER._test_category()   # indirect doctest
        """
        return None

    def __reduce__(self):
        r"""
        Used in pickling.

        TESTS::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: loads(dumps(ER)) == ER
            True
        """
        return CubicHeckeExtensionRing, (self._names, self._order, self._ring_of_definition)

    def _element_constructor_(self, x, mon=None):
        r"""
        Inherited element constructor overloaded to allow construction from
        ``GAP3`` ``MVP`` expressions.

        EXAMPLES::

            sage: # optional - gap3
            sage: CHA3 = algebras.CubicHecke(3)
            sage: GER = CHA3.extension_ring(generic=True)
            sage: sch7 = CHA3.chevie().SchurElements()[7]
            sage: GER(sch7)
            a*b*c^-2 + a^2*b^-1*c^-1 + a^-1*b^2*c^-1 + 2
            + a*b^-2*c + a^-2*b*c + a^-1*b^-1*c^2
            sage: rep4_gap3 = CHA3.chevie().Representations(4)
            sage: matrix(GER, rep4_gap3[1])
            [ b  0]
            [-b  c]
        """
        from sage.interfaces.gap3 import GAP3Element
        if isinstance(x, GAP3Element):
            return self._convert_from_gap3_mvp(x)
        return super()._element_constructor_(x, mon=mon)

    def _coerce_map_from_(self, R):
        r"""
        The rings that canonically coerce to ``self`` ar the ones from
        inheritance and the base ring of definition of the cubic Hecke algebra.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: ER = BR.extension_ring()
            sage: ER(BR.an_element())
            a*b + a*c + b*c + a*b^-1*c^-1 + 2*c^-1 + a^-1*b*c^-1 + 2*b^-1
            + 2*a^-1 + a^-1*b^-1*c
            sage: MBR = chbr.CubicHeckeRingOfDefinition(markov_trace_version=True)
            sage: MER = MBR.extension_ring()
            sage: MER(MBR.an_element())
            a*b*s^-1 + a*c*s^-1 + b*c*s^-1 + a*b^-1*c^-1 + 2*c^-1
            + a^-1*b*c^-1 + 2*b^-1 + 2*a^-1 + a^-1*b^-1*c
        """
        if isinstance(R, CubicHeckeRingOfDefinition):
            markov = R.markov_trace_version()
            a, b, c, *rem = self.gens()
            iu = a + b + c
            iv = a * b + a * c + b * c
            iw = a * b * c
            im_gens = [iu, iv, iw]
            if markov:
                if self.markov_trace_version():
                    im_gens += rem
                    # check of embedding fails in this case as long as the images of
                    # ``iu`` and ``iv`` need to be invertible (see comment in
                    # :meth:`__init__`).  # :class:`CubicHeckeRingOfDefinition`).
                    embedding_into_extension_ring = R.hom(im_gens, check=False)
            else:
                embedding_into_extension_ring = R.hom(im_gens)
            return embedding_into_extension_ring
        return super()._coerce_map_from_(R)

    def hom(self, im_gens, codomain=None, check=True, base_map=None):
        r"""
        Return a homomorphism of ``self``.

        INPUT:

        - ``im_gens`` -- tuple for the image of the generators of ``self``
        - ``codomain`` -- (optional) the codomain of the homomorphism

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: UCF = UniversalCyclotomicField()
            sage: map = ER.hom((UCF.gen(3),) + (UCF(3),UCF(4),UCF(5)))
            sage: ER.an_element()
            b^2*c^-1 + e3*a
            sage: map(_)
            -1/5*E(3) - 16/5*E(3)^2
        """
        gens = self.gens()
        num_gens = len(gens)

        if not isinstance(im_gens, (list, tuple)):
            im_gens = [im_gens]

        if len(im_gens) == num_gens + 1:
            e3, *im_remain = im_gens
            hom_cycl_gen = self.base_ring().hom([e3], codomain=e3.parent(), check=check, base_map=base_map)
            verbose("hom_cycl_gen %s" % hom_cycl_gen, level=2)
            return super().hom(im_remain, codomain=codomain, check=check, base_map=hom_cycl_gen)
        else:
            if base_map is None:
                raise ValueError('number of images must be four (including a '
                                 'third root of unity at first position) or a '
                                 'base_map (on %s) must be given' % self.base_ring())
            return super().hom(im_gens, codomain=codomain, check=check, base_map=base_map)

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('x, y, z')
            sage: ER.an_element()                             # indirect doctest
            y^2*z^-1 + e3*x
            sage: MER = chbr.CubicHeckeExtensionRing('x, y, z, s', markov_trace_version=True)
            sage: MER.an_element()                            # indirect doctest
            y^2*z^-1 + e3*x*s^-1
        """
        a, b, c, *rem = self.gens()
        e3 = self.cyclotomic_generator()
        s = self.one()
        if rem:
            s = rem[0]
        return b**2/c + a*e3/s

    ############################################################################
    # local methods
    ############################################################################
    def _is_markov_trace_version(self) -> bool:
        r"""
        Return whether ``self`` is the version containing the writhe parameter
        ``s`` for the Markov trace.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: ER._is_markov_trace_version()
            False
            sage: MER = chbr.CubicHeckeExtensionRing('a, b, c, s', markov_trace_version=True)
            sage: MER._is_markov_trace_version()
            True
        """
        return len(self.gens()) == 4

    # --------------------------------------------------------------------------
    # helper for element construction
    # --------------------------------------------------------------------------
    def _convert_from_gap3_mvp(self, mvp_expression):
        r"""
        Convert an element of ``GAP3`` interface containing Jean
        Michel's ``MVP`` (multivariate polynomials) to an element of ``self``.

        INPUT:

        - ``mvp_expression``  -- element of ``GAP3`` interface containing
          Jean Michel's ``MVP`` (multivariate polynomials)

        EXAMPLES::

            sage: # optional - gap3
            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: CHA3 = algebras.CubicHecke(3)
            sage: sch7 = CHA3.chevie().SchurElements()[7]
            sage: ER = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: ER._convert_from_gap3_mvp(sch7)
            a*b*c^-2 + a^2*b^-1*c^-1 + a^-1*b^2*c^-1 + 2 + a*b^-2*c + a^-2*b*c + a^-1*b^-1*c^2
        """
        from sage.misc.sage_eval import sage_eval
        E3 = self.cyclotomic_generator()
        a, b, c, *rem = self.gens()
        na, nb, nc = self.variable_names()
        lc = {na: a, nb: b, nc: c, 'E3': E3}
        sage_expr = str(mvp_expression.FormatMaple())[1:-1]
        return sage_eval(sage_expr, locals=lc)

    ############################################################################
    # global methods
    ############################################################################
    def cyclotomic_generator(self):
        r"""
        Return the third root of unity as generator of the base ring
        of ``self``.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER  = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: ER.cyclotomic_generator()
            e3
            sage: _**3 == 1
            True
        """
        return self(self.base_ring().gen())

    def conjugation(self):
        r"""
        Return an involution that performs *complex conjugation* with respect
        to base ring considered as order in the complex field.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('x, y, z')
            sage: conj = ER.conjugation()
            sage: conj(ER.an_element())
            y^2*z^-1 + (-e3 - 1)*x
            sage: MER = chbr.CubicHeckeExtensionRing('x, y, z, s', markov_trace_version=True)
            sage: conj = MER.conjugation()
            sage: conj(MER.an_element())
            y^2*z^-1 + (-e3 - 1)*x*s^-1
        """
        e3 = self.cyclotomic_generator()
        return self.hom(tuple([e3**2] + list(self.gens())))

    def cubic_equation_galois_group(self):
        r"""
        Return the Galois group of the cubic equation, which is the permutation
        group on the three generators together with its action on ``self``.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: G = ER.cubic_equation_galois_group()
            sage: t = ER.an_element()
            sage: [(g ,g*t) for g in G]
            [((), b^2*c^-1 + e3*a),
            ((1,3,2), a^2*b^-1 + e3*c),
            ((1,2,3), e3*b + a^-1*c^2),
            ((2,3), e3*a + b^-1*c^2),
            ((1,3), a^-1*b^2 + e3*c),
            ((1,2), a^2*c^-1 + e3*b)]
        """
        return self._galois_group

    def mirror_involution(self):
        r"""
        Return the involution of ``self`` corresponding to the involution of
        the cubic Hecke algebra (with the same name).

        This means that it maps the generators of ``self`` to their inverses.

        .. NOTE::

           The mirror involution of the braid group does not factor through the
           cubic Hecke algebra over its base ring, but it does if it is
           considered as `\ZZ`-algebra. The base ring elements are transformed
           by this automorphism.

        OUTPUT: the involution as automorphism of ``self``

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('p, q, r')
            sage: ER.mirror_involution()
            Ring endomorphism of Multivariate Laurent Polynomial Ring in p, q, r
                                 over Splitting Algebra of x^2 + x + 1
                                   with roots [e3, -e3 - 1]
                                 over Integer Ring
              Defn: p |--> p^-1
                    q |--> q^-1
                    r |--> r^-1
                    with map of base ring
            sage: _(ER.an_element())
            e3*p^-1 + q^-2*r

            sage: MER = chbr.CubicHeckeExtensionRing('p, q, r, s', markov_trace_version=True)
            sage: MER.mirror_involution()
            Ring endomorphism of Multivariate Laurent Polynomial Ring in p, q, r, s
              over Splitting Algebra of x^2 + x + 1
                with roots [e3, -e3 - 1] over Integer Ring
            Defn: p |--> p^-1
            q |--> q^-1
            r |--> r^-1
            s |--> s^-1
            with map of base ring
            sage: _(MER.an_element())
            e3*p^-1*s + q^-2*r
        """
        if self._mirror is None:
            e3 = self.base_ring().gen()
            if self._is_markov_trace_version():
                a, b, c, s = self.gens()
                self._mirror = self.hom([e3, ~a, ~b, ~c, ~s])
            else:
                a, b, c = self.gens()
                self._mirror = self.hom([e3, ~a, ~b, ~c])

        return self._mirror

    def create_specialization(self, im_cubic_equation_roots, im_writhe_parameter=None, var='T', third_unity_root_name='E3'):
        r"""
        Return an appropriate ring containing the elements from the list
        ``im_cubic_equation_roots`` defining a conversion map from ``self`` mapping
        the cubic equation roots of ``self`` to ``im_cubic_equation_roots``.

        INPUT:

        - ``im_cubic_equation_roots`` -- list or tuple of three ring elements
          such that there exists a ring homomorphism from the corresponding
          elements of ``self`` to them

        OUTPUT:

        A common parent containing the elements of ``im_cubic_equation_roots``
        together with their inverses.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: t = ER.an_element(); t
            b^2*c^-1 + e3*a
            sage: Sp1 = ER.create_specialization([E(5), E(7), E(3)]); Sp1
            Universal Cyclotomic Field
            sage: Sp1(t)
            -E(105)^11 - E(105)^16 - E(105)^26 - E(105)^37 - E(105)^41
            - E(105)^58 - E(105)^71 - E(105)^79 - E(105)^86 - E(105)^101
            sage: MER = chbr.CubicHeckeExtensionRing('a, b, c, s', markov_trace_version=True)
            sage: MER.create_specialization([E(5), E(7), E(3)], im_writhe_parameter=E(4))
            Universal Cyclotomic Field
            sage: a, b, c, s = MER.gens()
            sage: Sp1(MER(t)/s)
            E(420) + E(420)^29 + E(420)^89 + E(420)^149 + E(420)^169 + E(420)^209
            + E(420)^253 + E(420)^269 + E(420)^337 + E(420)^389

            sage: Z3 = CyclotomicField(3); E3=Z3.gen()
            sage: Sp2 = ER.create_specialization([E3, E3**2, Z3(1)])
            sage: Sp2(t)
            -1
            sage: MER.create_specialization([E3, E3**2, 1], im_writhe_parameter=2)
            Cyclotomic Field of order 3 and degree 2
            sage: Sp2(MER(t)*s)
            -2

            sage: Sp3 = ER.create_specialization([5, 7, 11])
            sage: Sp3(t)
            5*E3 + 49/11
        """
        # ----------------------------------------------------------------------
        # interpreting user given cubic equation roots and define the
        # corresponding specialized extension ring.
        # ----------------------------------------------------------------------

        if isinstance(im_cubic_equation_roots, tuple):
            im_cubic_equation_roots = list(im_cubic_equation_roots)

        if not isinstance(im_cubic_equation_roots, list):
            raise TypeError('cubic_equation_roots must be a list of three elements')

        if len(im_cubic_equation_roots) != 3:
            raise ValueError('there must be exactly three cubic_equation_roots')

        gens = self.gens()
        num_gens = len(gens)
        if im_writhe_parameter:
            if num_gens < 4:
                raise ValueError('im_writhe_parameter only possible for Markov-trace extension')
            im_gens = im_cubic_equation_roots + [im_writhe_parameter]
            a, b, c, s = im_gens
        else:
            if num_gens == 4:
                raise ValueError('im_writhe_parameter must be given for Markov-trace extension')
            im_gens = im_cubic_equation_roots
            a, b, c = im_gens

        image_ring = get_coercion_model().common_parent(*(im_gens))

        # ----------------------------------------------------------------------
        # make sure that all given cubic equation roots and their inverses
        # belong to image_ring
        # ----------------------------------------------------------------------
        try:
            image_ring = image_ring.localization(tuple(im_gens))
        except ValueError:
            pass

        im_gens = [image_ring(root) for root in im_gens]
        verbose('common parent of roots and inverses: %s' % (image_ring), level=2)

        image_ring_base = image_ring.base_ring()
        image_ring_map = None

        verbose('first choice: image_ring %s, image_ring_base %s' % (image_ring, image_ring_base), level=2)

        # ----------------------------------------------------------------------
        # make sure that a third root of unity belongs to image_ring
        # ----------------------------------------------------------------------

        E3 = None
        cp3 = cyclotomic_polynomial(3, var=var).change_ring(image_ring)
        cyclotomic_roots = solve_with_extension(cp3, [third_unity_root_name], var=var, flatten=True, warning=False)

        if len(cyclotomic_roots) > 0:
            E3 = cyclotomic_roots[0]
            verbose('third root of unity %s found in %s' % (E3, E3.parent()), level=2)

        if E3 is None:
            raise RuntimeError('cannot find a ring containing a third root of unity for the this choice of cubic roots!')

        hom_gens = [E3] + im_gens
        verbose('hom_gens %s' % hom_gens, level=2)

        image_ring = get_coercion_model().common_parent(*(hom_gens))
        verbose('common parent of roots and third root: %s' % image_ring, level=2)

        hom_gens = [image_ring(gen) for gen in hom_gens]

        image_ring_base = image_ring.base_ring()

        verbose('second choice: image_ring %s, image_ring_base %s' % (image_ring, image_ring_base), level=2)

        try:
            image_ring_map = self.hom(hom_gens, codomain=image_ring)
        except (ValueError, NotImplementedError):
            image_ring_map = self.hom(hom_gens, codomain=image_ring, check=False)
            verbose('check failed for embedding as ring morphism')

        verbose('specializing map defined %s' % image_ring_map, level=2)

        register_ring_hom(image_ring_map)
        return image_ring

    def as_splitting_algebra(self):
        r"""
        Return ``self`` as a :class:`SplittingAlgebra`; that is as an
        extension ring of the corresponding cubic Hecke algebra base ring
        (``self._ring_of_definition``, as a :class:`CubicHeckeRingOfDefinition`)
        splitting its cubic equation into linear factors, such that the roots
        are images of the generators of ``self``.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: GBR = chbr.CubicHeckeRingOfDefinition()
            sage: GER = GBR.extension_ring()
            sage: ER = GER.as_splitting_algebra(); ER
            Splitting Algebra of T^2 + T + 1 with roots [E3, -E3 - 1]
              over Splitting Algebra of h^3 - u*h^2 + v*h - w
                with roots [a, b, -b - a + u]
              over Multivariate Polynomial Ring in u, v, w
              over Integer Ring localized at (w,)
            sage: ER(GER.an_element())
            a*E3 + ((u/(-w))*a^2 + ((u^2 - v)/w)*a)*b + a - u
            sage: ER(GBR.an_element())
            (u^2 + v*w)/w

            sage: MBR = chbr.CubicHeckeRingOfDefinition(markov_trace_version=True)
            sage: MER = MBR.extension_ring()
            sage: ES = MER.as_splitting_algebra(); ES
            Splitting Algebra of T^2 + T + 1 with roots [E3, -E3 - 1]
              over Splitting Algebra of h^3 - u*h^2 + v*h - w
                with roots [a, b, -b - a + u]
              over Multivariate Polynomial Ring in u, v, w, s
              over Integer Ring localized at (s, w, v, u)
            sage: ES(MER.an_element())
            (((-1)/(-s))*a)*E3 + ((u/(-w))*a^2 + ((u^2 - v)/w)*a)*b + a - u
            sage: ES(MBR.an_element())
            (u^2*s + v*w)/(w*s)
        """
        if self._splitting_algebra is not None:
            verbose("End (short)", level=2)
            return self._splitting_algebra

        if self._ring_of_definition is None:
            verbose("constructing generic base ring", level=2)
            self._ring_of_definition = CubicHeckeRingOfDefinition()

        markov = self._is_markov_trace_version()

        BR = self._ring_of_definition
        root_names = list(self._names)
        var_s = None
        if markov:
            var_s = root_names.pop()  # s not needed as root
            a, b, c, s = self.gens()
        else:
            a, b, c = self.gens()

        root_names.pop()  # c not needed as root

        FSR = SplittingAlgebra(BR.cubic_equation(), root_names, warning=False)
        splitting_roots = FSR.splitting_roots()
        verbose('splitting roots %s' % splitting_roots, level=2)

        A, B, C = splitting_roots
        e3 = self.cyclotomic_generator()
        if var_s:
            fsr_s = FSR.scalar_base_ring().gens_dict()[var_s]
            S = self.create_specialization([A, B, C], im_writhe_parameter=fsr_s)
            # check of embedding fails in this case as long as the images of
            # ``iu`` and ``iv`` need to be invertible (see comment in
            # :meth:`__init__` of :class:`CubicHeckeRingOfDefinition`).
            map_back = S.hom([e3, b, a, a + b + c, a*b+a*c+b*c, a*b*c, s], check=False)
        else:
            S = self.create_specialization([A, B, C])
            map_back = S.hom([e3, b, a, a + b + c, a*b+a*c+b*c, a*b*c])
        self.register_coercion(map_back)
        self._splitting_algebra = S
        return self._splitting_algebra

    def field_embedding(self, characteristic=0):
        r"""
        Return a field embedding of ``self``.

        INPUT:

        - ``characteristic`` -- integer (default: `0`); the characteristic
          of the field

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: ER = BR.extension_ring()
            sage: ER.field_embedding()
            Ring morphism:
            From: Multivariate Laurent Polynomial Ring in a, b, c
                    over Splitting Algebra of x^2 + x + 1
                      with roots [e3, -e3 - 1]
                    over Integer Ring
            To:   Fraction Field of Multivariate Polynomial Ring in a, b, c
                    over Cyclotomic Field of order 3 and degree 2
            Defn: a |--> a
                  b |--> b
                  c |--> c
            with map of base ring

            sage: ER.field_embedding(characteristic=5)
            Ring morphism:
            From: Multivariate Laurent Polynomial Ring in a, b, c
                    over Splitting Algebra of x^2 + x + 1
                      with roots [e3, -e3 - 1]
                    over Integer Ring
            To:   Fraction Field of Multivariate Polynomial Ring in a, b, c
                    over Finite Field in a of size 5^2
            Defn: a |--> a
                  b |--> b
                  c |--> c
            with map of base ring

            sage: MER = ER.markov_trace_version()
            sage: MER.field_embedding()
            Ring morphism:
            From: Multivariate Laurent Polynomial Ring in a, b, c, s
                    over Splitting Algebra of x^2 + x + 1
                      with roots [e3, -e3 - 1]
                    over Integer Ring
            To:   Fraction Field of Multivariate Polynomial Ring in a, b, c, s
                    over Cyclotomic Field of order 3 and degree 2
            Defn: a |--> a
                  b |--> b
                  c |--> c
                  s |--> s
            with map of base ring
        """
        if characteristic == 0:
            from sage.rings.number_field.number_field import CyclotomicField
            C3 = CyclotomicField(3)
            E3 = C3.gen()
        else:
            if not ZZ(characteristic).is_prime():
                raise ValueError('characteristic must be a prime integer')
            from sage.rings.finite_rings.finite_field_constructor import GF
            from sage.misc.functional import cyclotomic_polynomial
            G = GF(characteristic)
            c3 = cyclotomic_polynomial(3).change_ring(G)
            C3 = c3.splitting_field('a')
            E3 = c3.change_ring(C3).roots()[0][0]

        B = self.base_ring()
        embBase = B.hom((E3,))
        if not C3.has_coerce_map_from(B):
            C3._unset_coercions_used()
            C3.register_coercion(embBase)
        P = C3[self.variable_names()]
        F = P.fraction_field()
        emb = self.hom((F(E3),) + F.gens())
        F = emb.codomain()
        if not F.has_coerce_map_from(self):
            F._unset_coercions_used()
            F.register_coercion(emb)
        return emb

    def markov_trace_version(self):
        r"""
        Return the Markov trace version of ``self``.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: ER = chbr.CubicHeckeExtensionRing('a, b, c')
            sage: ER.markov_trace_version()
            Multivariate Laurent Polynomial Ring in a, b, c, s
              over Splitting Algebra of x^2 + x + 1
                with roots [e3, -e3 - 1] over Integer Ring
        """
        if self._is_markov_trace_version():
            return self
        names = self.variable_names() + ('s',)
        return self.__class__(names=names, order=self._order, markov_trace_version=True)


################################################################################
# Ring of Definition
# ------------------------------------------------------------------------------
# Definition of the ring of definition for the cubic Hecke algebra as polynomial
# ring in 2 indeterminates over univariate Laurent polynomial ring over the
# integers. This is the most general ring over which the cubic Hecke algebra may
# be defined.
# ------------------------------------------------------------------------------
class CubicHeckeRingOfDefinition(Localization):
    r"""
    The *ring of definition* of the cubic Hecke algebra.

    It contains one invertible indeterminate (representing the product of the
    roots of the cubic equation) and two non invertible indeterminates.

    .. NOTE::

        We follow a suggestion by Ivan Marin in the name *ring of definition*.
        We avoid alternative names like *generic* or *universal* base ring
        as these have some issues. The first option could be misleading
        in view of the term *generic point* used in algebraic geometry, which
        would mean the function field in ``u, v, w``, here.

        The second option is problematic since the base ring itself is not a
        universal object. Rather, the universal object is the cubic Hecke algebra
        considered as a `\ZZ`-algebra including ``u, v, w`` as pairwise commuting
        indeterminates. From this point of view the base ring appears to be a
        subalgebra of this universal object generated by ``u, v, w``.

    INPUT:

    - ``names`` -- (default: ``'u,v,w'``) string containing the names of the
      indeterminates separated by ``,`` or a triple of strings each of which
      are the names of one of the three indeterminates
    - ``order`` -- string (default: ``'degrevlex'``); the term order; see also
      :class:`~sage.rings.polynomial.laurent_polynomial_ring.LaurentPolynomialRing_mpair`
    - ``ring_of_definition`` -- (optional) a :class:`CubicHeckeRingOfDefinition`
      to specify the generic cubic Hecke base ring over which ``self`` may be
      realized as splitting ring via the ``as_splitting_algebra`` method
    - ``markov_trace_version`` -- boolean (default: ``False``); if this is
      set to ``True`` then ``self`` contains one invertible indeterminate in
      addition which is meant to represent the writhe factor of a Markov trace
      on the cubic Hecke algebra and which default name is ``s``

    EXAMPLES::

        sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
        sage: BR = chbr.CubicHeckeRingOfDefinition()
        sage: u, v, w = BR.gens()
        sage: ele = 3*u*v-5*w**(-2)
        sage: ER = BR.extension_ring()
        sage: ER(ele)
        3*a^2*b + 3*a*b^2 + 3*a^2*c + 9*a*b*c + 3*b^2*c
        + 3*a*c^2 + 3*b*c^2 + (-5)*a^-2*b^-2*c^-2
        sage: phi1 = BR.hom( [4,3,1/1] )
        sage: phi1(ele)
        31

        sage: LL.<t> = LaurentPolynomialRing(ZZ)
        sage: phi2=BR.hom( [LL(4),LL(3),t] )
        sage: phi2(ele)
        -5*t^-2 + 36

        sage: BR.create_specialization( [E(5), E(7), E(3)] )
        Universal Cyclotomic Field
        sage: _(ele)
        -3*E(105) - 5*E(105)^2 - 5*E(105)^8 - 5*E(105)^11 - 5*E(105)^17
        - 5*E(105)^23 - 5*E(105)^26 - 5*E(105)^29 - 5*E(105)^32 - 5*E(105)^38
        - 5*E(105)^41 - 5*E(105)^44 - 5*E(105)^47 - 5*E(105)^53 - 5*E(105)^59
        - 5*E(105)^62 - 5*E(105)^68 - 8*E(105)^71 - 5*E(105)^74 - 5*E(105)^83
        - 5*E(105)^86 - 5*E(105)^89 - 5*E(105)^92 - 5*E(105)^101 - 5*E(105)^104
    """
    def __init__(self, names=('u', 'v', 'w', 's'), order='degrevlex', markov_trace_version=False):
        r"""
        Initialize ``self``.

        TESTS::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: TestSuite(BR).run()
        """
        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------
        # Saving class-globals
        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------
        names = normalize_names_markov(names, markov_trace_version)
        if len(names) == 4:
            # invertible_names = names[2:4]  # s must be invertible, too
            #
            # because the formal Markov trace of the both basis elements
            # ``self.get_order()[568]`` (``KnotInfo.L8a7_1``) and
            # ``self.get_order()[596]`` (mirror image of ``KnotInfo.K9_46``)
            # have the indeterminate ``v`` in their denominator we need to have
            # all indeterminates invertible (``u`` as well for the mirror images)
            invertible_names = names
        else:
            invertible_names = names[2]

        self._order = order

        # ----------------------------------------------------------------------
        # ---------------------------------------------------------------------
        # Init of self
        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------
        base_ring = PolynomialRing(ZZ, names, order=order)
        Localization.__init__(self, base_ring, invertible_names)

        # ----------------------------------------------------------------------
        # Init of data used on demand
        # ----------------------------------------------------------------------
        self._mirror = None

    # ########################################################################
    # overloaded inherited methods
    # ########################################################################
    def _defining_names(self):
        r"""
        Return the generators of ``self`` as the defining names.

        This method is cached in the parent class.
        This causes trouble if a second instance of ``self`` is used for another
        cubic Hecke algebra in the same session. To avoid this it is overloaded
        without ``cached_method`` decorator.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: BR._defining_names()
            (u, v, w)
        """
        return self.gens()

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: BR.an_element()                            # indirect doctest
            (u^2 + v*w)/w
            sage: MBR = chbr.CubicHeckeRingOfDefinition(markov_trace_version=True)
            sage: MBR.an_element()                           # indirect doctest
            (u^2*s + v*w)/(w*s)
        """
        u, v, w, *rem = self.gens()
        s = self.one()
        if rem:
            s = rem[0]
        return u**2/w+v/s

    ############################################################################
    # Local Methods
    ############################################################################
    def _is_markov_trace_version(self) -> bool:
        r"""
        Return whether ``self`` is the version containing the writhe parameter
        ``s`` for the Markov trace.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: BR._is_markov_trace_version()
            False
            sage: MBR = chbr.CubicHeckeRingOfDefinition(markov_trace_version=True)
            sage: MBR._is_markov_trace_version()
            True
        """
        return len(self.gens()) == 4

    ############################################################################
    # Global Methods
    ############################################################################
    def cubic_equation(self, var='h', as_coefficients=False):
        r"""
        Return the cubic equation over which the cubic Hecke algebra is defined.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: BR.cubic_equation()
            h^3 - u*h^2 + v*h - w
            sage: BR.cubic_equation(var='t')
            t^3 - u*t^2 + v*t - w
            sage: BR.cubic_equation(as_coefficients=True)
            [-w, v, -u, 1]
        """
        u, v, w, *rem = self.gens()
        cf = [-w, v, -u, 1]
        if as_coefficients:
            return cf
        P = PolynomialRing(self, var)

        return P(cf)

    def mirror_involution(self):
        r"""
        Return the involution of ``self`` corresponding to the involution of the
        cubic Hecke algebra (with the same name).

        This means that it maps the last generator of ``self`` to its inverse
        and both others to their product with the image of the former.

        From the cubic equation for a braid generator `\beta_i`:

        .. MATH::

            \beta_i^3 - u \beta_i^2 + v\beta_i -w = 0.

        One deduces the following cubic equation for `\beta_i^{-1}`:

        .. MATH::

            \beta_i^{-3} -\frac{v}{w} \beta_i^{-2} + \frac{u}{w}\beta_i^{-1}
            - \frac{1}{w} = 0.

        .. NOTE::

           The mirror involution of the braid group does not factor through
           the cubic Hecke algebra over its base ring, but it does if it is
           considered as `\ZZ`-algebra. The base ring elements are transformed
           by this automorphism.

        OUTPUT: the involution as automorphism of ``self``

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: BR.mirror_involution()
            Ring endomorphism of Multivariate Polynomial Ring in u, v, w
                                 over Integer Ring localized at (w,)
              Defn: u |--> v/w
                    v |--> u/w
                    w |--> 1/w
            sage: _(BR.an_element())
            (v^2 + u)/w

            sage: MBR = chbr.CubicHeckeRingOfDefinition(markov_trace_version=True)
            sage: MBR.mirror_involution()
            Ring endomorphism of Multivariate Polynomial Ring in u, v, w, s
                                 over Integer Ring localized at (s, w, v, u)
            Defn: u |--> v/w
            v |--> u/w
            w |--> 1/w
            s |--> 1/s
            sage: _(MBR.an_element())
            (v^2 + u*s)/w
        """
        if self._mirror is None:
            if self._is_markov_trace_version():
                u, v, w, s = self.gens()
                self._mirror = self.hom([v / w, u / w, ~w, ~s])
            else:
                u, v, w = self.gens()
                self._mirror = self.hom([v / w, u / w, ~w])
        return self._mirror

    def create_specialization(self, im_cubic_equation_parameters, im_writhe_parameter=None):
        r"""
        Return an appropriate Ring containing the elements from the list
        ``im_cubic_equation_parameters`` having a conversion map from ``self``
        mapping the cubic equation parameters of ``self`` to
        ``im_cubic_equation_parameters``.

        INPUT:

        - ``im_cubic_equation_parameters`` -- list or tuple of three ring
          elements such that there exists a ring homomorphism from the
          corresponding elements of ``self`` to them

        OUTPUT:

        A common parent containing the elements of ``im_cubic_equation_parameters``
        together with an inverse of the third element.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: t = BR.an_element(); t
            (u^2 + v*w)/w
            sage: Sp1 = BR.create_specialization([E(5), E(7), E(3)]); Sp1
            Universal Cyclotomic Field
            sage: Sp1(t)
            E(105) + E(105)^8 + E(105)^29 - E(105)^37 + E(105)^43 - E(105)^52
            + E(105)^64 - E(105)^67 + E(105)^71 - E(105)^82 + E(105)^92
            - E(105)^97

            sage: MBR = chbr.CubicHeckeRingOfDefinition(markov_trace_version=True)
            sage: MBR.create_specialization([E(5), E(7), E(3)], im_writhe_parameter=E(4))
            Universal Cyclotomic Field
            sage: u, v, w, s = MBR.gens()
            sage: Sp1(MBR(t)/s)
            E(420)^13 - E(420)^53 + E(420)^73 - E(420)^109 - E(420)^137
            - E(420)^221 + E(420)^253 - E(420)^277 + E(420)^313 - E(420)^361
            + E(420)^373 - E(420)^389

            sage: Z3 = CyclotomicField(3); E3=Z3.gen()
            sage: Sp2 = BR.create_specialization([E3, E3**2, 1]); Sp2
            Cyclotomic Field of order 3 and degree 2
            sage: Sp2(t)
            -2*zeta3 - 2
            sage: MBR.create_specialization([E3, E3**2, 1], im_writhe_parameter=2)
            Cyclotomic Field of order 3 and degree 2
            sage: Sp2(MBR(t)/s)
            -zeta3 - 1

            sage: Sp3 = BR.create_specialization([5, 7, 11]); Sp3
            Integer Ring localized at (11,)
            sage: Sp3(t)
            102/11
        """
        # ----------------------------------------------------------------------
        # setting the base_ring  according to the cubic_equation_parameters
        # ----------------------------------------------------------------------
        if isinstance(im_cubic_equation_parameters, tuple):
            im_cubic_equation_parameters = list(im_cubic_equation_parameters)

        if not isinstance(im_cubic_equation_parameters, list):
            raise TypeError('cubic_equation_parameters must be a list of three elements')

        if len(im_cubic_equation_parameters) != 3:
            raise ValueError('there must be exactly three cubic_equation_parameters')

        gens = self.gens()
        num_gens = len(gens)
        if im_writhe_parameter:
            if num_gens < 4:
                raise ValueError('im_writhe_parameter only possible for Markov-trace extension')
            im_gens = im_cubic_equation_parameters + [im_writhe_parameter]
            u, v, w, s = im_gens
        else:
            if num_gens == 4:
                raise ValueError('im_writhe_parameter must be given for Markov-trace extension')
            im_gens = im_cubic_equation_parameters
            u, v, w = im_gens

        image_ring = None
        image_ring_map = None
        image_ring_base = w.parent()

        # ----------------------------------------------------------------------
        # short exit on trivial invocation
        # ----------------------------------------------------------------------
        if image_ring_base is self and im_gens == gens:
            return self

        image_ring = get_coercion_model().common_parent(*(im_gens))

        # ----------------------------------------------------------------------
        # make sure that the inverse of w belongs to image_ring
        # ----------------------------------------------------------------------
        try:
            image_ring = image_ring.localization(w)
        except ValueError:
            pass

        im_gens = [image_ring(para) for para in im_gens]

        verbose('common parent of parameters and inverses: %s' % image_ring, level=2)

        try:
            image_ring_map = self.hom(im_gens, codomain=image_ring)
        except ValueError:
            image_ring_map = self.hom(im_gens, codomain=image_ring, check=False)
            verbose('Warning: check failed for embedding as ring morphism')

        register_ring_hom(image_ring_map)
        return image_ring

    # --------------------------------------------------------------------------
    # Definition of the generic extension ring for the cubic hecke algebra as
    # Laurent polynomial ring in 3 indeterminates over cyclotomic field of order
    # 3. The generic extension ring guarantees semisimplicity of the cubic Hecke
    # algebra.
    # --------------------------------------------------------------------------
    @cached_method
    def extension_ring(self, names=('a', 'b', 'c', 's')):
        r"""
        Return the generic extension ring attached to ``self``.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: BR = chbr.CubicHeckeRingOfDefinition()
            sage: BR.extension_ring()
            Multivariate Laurent Polynomial Ring in a, b, c
              over Splitting Algebra of x^2 + x + 1
                with roots [e3, -e3 - 1]
              over Integer Ring
        """
        markov = self._is_markov_trace_version()
        return CubicHeckeExtensionRing(names, ring_of_definition=self, markov_trace_version=markov)

    def markov_trace_version(self):
        r"""
        Return the extension of the ring of definition needed to treat the
        formal Markov traces.

        This appends an additional variable ``s`` to measure the writhe
        of knots and makes ``u`` and ``v`` invertible.

        EXAMPLES::

            sage: from sage.algebras.hecke_algebras import cubic_hecke_base_ring as chbr
            sage: GBR = chbr.CubicHeckeRingOfDefinition()
            sage: GBR.markov_trace_version()
            Multivariate Polynomial Ring in u, v, w, s
              over Integer Ring localized at (s, w, v, u)
        """
        if self._is_markov_trace_version():
            return self
        names = self.base_ring().variable_names() + ('s',)
        return self.__class__(names=names, order=self._order, markov_trace_version=True)

    def specialize_homfly(self):
        r"""
        Return a map to the two variable Laurent polynomial ring that is
        the parent of the HOMFLY-PT polynomial.

        EXAMPLES::

            sage: from sage.knots.knotinfo import KnotInfo
            sage: CHA2 = algebras.CubicHecke(2)
            sage: K5_1 = KnotInfo.K5_1.link()
            sage: br = CHA2(K5_1.braid())
            sage: mt = br.formal_markov_trace()
            sage: MT = mt.base_ring()
            sage: f = MT.specialize_homfly(); f
            Composite map:
              From: Multivariate Polynomial Ring in u, v, w, s over Integer Ring
                    localized at (s, w, v, u)
              To:   Multivariate Laurent Polynomial Ring in L, M over Integer Ring
              Defn:   Ring morphism:
                      From: Multivariate Polynomial Ring in u, v, w, s
                            over Integer Ring localized at (s, w, v, u)
                      To:   Multivariate Polynomial Ring in L, M
                            over Integer Ring localized at (M - 1, M, L)
                      Defn: u |--> -M + 1
                            v |--> -M + 1
                            w |--> 1
                            s |--> L
                    then
                      Conversion map:
                      From: Multivariate Polynomial Ring in L, M
                            over Integer Ring localized at (M - 1, M, L)
                      To:   Multivariate Laurent Polynomial Ring in L, M
                            over Integer Ring
            sage: sup = mt.support()
            sage: h1 = sum(f(mt.coefficient(b)) * b.regular_homfly_polynomial() for b in sup)
            sage: L, M = f.codomain().gens()
            sage: h2 = K5_1.homfly_polynomial()
            sage: h1*L**(-5) == h2  # since the writhe of K5_1 is 5
            True
        """
        if not self._is_markov_trace_version():
            raise ValueError('functionality only available for Markov trace version')
        from sage.knots.link import Link
        H = Link([]).homfly_polynomial().parent()
        L, M = H.gens()
        HL = H.localization(1 - M)
        u = HL(1 - M)
        phi = self.hom((u, u, HL.one(), HL(L)))
        inc = H.convert_map_from(HL)
        return inc * phi

    def specialize_kauffman(self):
        r"""
        Return a map to the two variable Laurent polynomial ring that is
        the parent of the Kauffman polynomial.

        EXAMPLES::

            sage: from sage.knots.knotinfo import KnotInfo
            sage: CHA2 = algebras.CubicHecke(2)
            sage: K5_1 = KnotInfo.K5_1.link()
            sage: br = CHA2(K5_1.braid())
            sage: mt = br.formal_markov_trace()
            sage: MT = mt.base_ring()
            sage: f = MT.specialize_kauffman(); f
            Composite map:
              From: Multivariate Polynomial Ring in u, v, w, s over Integer Ring
                    localized at (s, w, v, u)
              To:   Multivariate Laurent Polynomial Ring in a, z over Integer Ring
              Defn:   Ring morphism:
                      From: Multivariate Polynomial Ring in u, v, w, s
                            over Integer Ring localized at (s, w, v, u)
                      To:   Multivariate Polynomial Ring in a, z
                            over Integer Ring localized at (z, a, a + z, a*z + 1)
                      Defn: u |--> (a*z + 1)/a
                            v |--> (a + z)/a
                            w |--> 1/a
                            s |--> a
                    then
                      Conversion map:
                      From: Multivariate Polynomial Ring in a, z over Integer Ring
                            localized at (z, a, a + z, a*z + 1)
                      To:   Multivariate Laurent Polynomial Ring in a, z
                            over Integer Ring
            sage: sup = mt.support()
            sage: k1 = sum(f(mt.coefficient(b)) * b.regular_kauffman_polynomial() for b in sup)
            sage: a, z = f.codomain().gens()
            sage: k2 = KnotInfo.K5_1.kauffman_polynomial()
            sage: k1*a**(-5) == k2  # since the writhe of K5_1 is 5
            True
        """
        if not self._is_markov_trace_version():
            raise ValueError('functionality only available for Markov trace version')
        from sage.knots.knotinfo import KnotInfo
        K = KnotInfo.L2a1_1.kauffman_polynomial().parent()
        a, z = K.gens()
        ku = z * a + 1
        kv = z + a
        KL = K.localization((ku, kv))
        u = KL(ku / a)
        v = KL(kv / a)
        phi = self.hom((u, v, KL(~a), KL(a)))
        inc = K.convert_map_from(KL)
        return inc * phi

    def specialize_links_gould(self):
        r"""
        Return a map to the two variable Laurent polynomial ring that is
        the parent of the Links-Gould polynomial.

        EXAMPLES::

            sage: from sage.knots.knotinfo import KnotInfo
            sage: CHA2 = algebras.CubicHecke(2)
            sage: K5_1 = KnotInfo.K5_1.link()
            sage: br = CHA2(K5_1.braid())
            sage: mt = br.formal_markov_trace()
            sage: MT = mt.base_ring()
            sage: f = MT.specialize_links_gould(); f
            Composite map:
              From: Multivariate Polynomial Ring in u, v, w, s over Integer Ring
                    localized at (s, w, v, u)
              To:   Multivariate Laurent Polynomial Ring in t0, t1 over Integer Ring
              Defn:   Ring morphism:
                      From: Multivariate Polynomial Ring in u, v, w, s
                            over Integer Ring localized at (s, w, v, u)
                      To:   Multivariate Polynomial Ring in t0, t1 over Integer Ring
                            localized at (t1, t0, t0 + t1 - 1, t0*t1 - t0 - t1)
                      Defn: u |--> t0 + t1 - 1
                            v |--> t0*t1 - t0 - t1
                            w |--> -t0*t1
                            s |--> 1
                    then
                      Conversion map:
                      From: Multivariate Polynomial Ring in t0, t1 over Integer Ring
                            localized at (t1, t0, t0 + t1 - 1, t0*t1 - t0 - t1)
                      To:   Multivariate Laurent Polynomial Ring in t0, t1 over Integer Ring
            sage: sup = mt.support()
            sage: sum(f(mt.coefficient(b)) * b.links_gould_polynomial() for b in sup)
            -t0^4*t1 - t0^3*t1^2 - t0^2*t1^3 - t0*t1^4 + t0^4 + 2*t0^3*t1 + 2*t0^2*t1^2
            + 2*t0*t1^3 + t1^4 - t0^3 - 2*t0^2*t1 - 2*t0*t1^2 - t1^3 + t0^2 + 2*t0*t1
            + t1^2 - t0 - t1 + 1
        """
        if not self._is_markov_trace_version():
            raise ValueError('functionality only available for Markov trace version')
        from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
        L = LaurentPolynomialRing(ZZ, 't0, t1')
        t0, t1 = L.gens()
        lu = t0 + t1 - 1
        lv = t0 * t1 - t0 - t1
        lw = -t0 * t1
        LL = L.localization((lu, lv))
        u = LL(lu)
        v = LL(lv)
        w = LL(lw)
        phi = self.hom((u, v, w, LL.one()))
        inc = L.convert_map_from(LL)
        return inc * phi



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules

from cpython.object cimport *
from cysignals.signals cimport sig_on, sig_off

cdef extern from 'symmetrica/def.h':
    INT kostka_number(OP shape, OP content, OP result)
    INT kostka_tab(OP shape, OP content, OP result)
    INT kostka_tafel(OP n, OP result)


def kostka_number_symmetrica(shape, content):
    """
    Compute the kostkanumber, i.e. the number of
    tableaux of given shape, which is a PARTITION object, and
    of given content, which also is a PARTITION object, or a VECTOR
    object with INTEGER entries. The
    result is an INTEGER object, which is freed to an empty
    object at the beginning. The shape could also be a
    SKEWPARTITION object, then we compute the number of
    skewtableaux of the given shape.

    EXAMPLES::

        sage: symmetrica.kostka_number([2,1],[1,1,1])
        2
        sage: symmetrica.kostka_number([1,1,1],[1,1,1])
        1
        sage: symmetrica.kostka_number([3],[1,1,1])
        1
    """
    cdef OP cshape = callocobject(), ccontent = callocobject(), result = callocobject()

    if isinstance(shape, <type>builtinlist):
        if isinstance(shape[0], <type>builtinlist):
            shape = SkewPartition(shape)
        else:
            shape = Partition(shape)


    if isinstance(shape, <type>SkewPartition):
        _op_skew_partition(shape, cshape)
    else:
        _op_partition(shape, cshape)

    _op_partition(content, ccontent)

    kostka_number(ccontent, cshape, result)

    res = _py(result)

    freeall(cshape)
    freeall(ccontent)
    freeall(result)


    return res


def kostka_tab_symmetrica(shape, content):
    """
    Compute the list of tableaux of given shape
    and content. shape is a PARTITION object or a
    SKEWPARTITION object and
    content is a PARTITION object or a VECTOR object with
    INTEGER entries, the result becomes a
    LIST object whose entries are the computed TABLEAUX
    object.

    EXAMPLES::

        sage: symmetrica.kostka_tab([3],[1,1,1])
        [[[1, 2, 3]]]
        sage: symmetrica.kostka_tab([2,1],[1,1,1])
        [[[1, 2], [3]], [[1, 3], [2]]]
        sage: symmetrica.kostka_tab([1,1,1],[1,1,1])
        [[[1], [2], [3]]]
        sage: symmetrica.kostka_tab([[2,2,1],[1,1]],[1,1,1])
        [[[None, 1], [None, 2], [3]],
         [[None, 1], [None, 3], [2]],
         [[None, 2], [None, 3], [1]]]
        sage: symmetrica.kostka_tab([[2,2],[1]],[1,1,1])
        [[[None, 1], [2, 3]], [[None, 2], [1, 3]]]
    """
    late_import()

    cdef OP cshape = callocobject(), ccontent = callocobject(), result = callocobject()
    cdef INT err

    if isinstance(shape, <type>builtinlist):
        if isinstance(shape[0], <type>builtinlist):
            shape = SkewPartition(shape)
        else:
            shape = Partition(shape)


    if isinstance(shape, <type>SkewPartition):
        _op_skew_partition(shape, cshape)
    else:
        _op_partition(shape, cshape)


    #Check to make sure the content is compatible with the shape.

    _op_il_vector(content, ccontent)

    err = kostka_tab(cshape, ccontent, result)

    res = _py(result)

    freeall(cshape)
    freeall(ccontent)
    freeall(result)


    return res


def kostka_tafel_symmetrica(n):
    """
    Return the table of Kostka numbers of weight `n`.

    EXAMPLES::

        sage: symmetrica.kostka_tafel(1)
        [1]

        sage: symmetrica.kostka_tafel(2)
        [1 0]
        [1 1]

        sage: symmetrica.kostka_tafel(3)
        [1 0 0]
        [1 1 0]
        [1 2 1]

        sage: symmetrica.kostka_tafel(4)
        [1 0 0 0 0]
        [1 1 0 0 0]
        [1 1 1 0 0]
        [1 2 1 1 0]
        [1 3 2 3 1]

        sage: symmetrica.kostka_tafel(5)
        [1 0 0 0 0 0 0]
        [1 1 0 0 0 0 0]
        [1 1 1 0 0 0 0]
        [1 2 1 1 0 0 0]
        [1 2 2 1 1 0 0]
        [1 3 3 3 2 1 0]
        [1 4 5 6 5 4 1]
    """

    cdef OP cn = callocobject(), cresult = callocobject()

    _op_integer(n, cn)

    sig_on()
    kostka_tafel(cn, cresult)
    sig_off()

    res = _py(cresult)

    freeall(cn)
    freeall(cresult)

    return res



# sage_setup: distribution = sagemath-combinat
r"""
Signed compositions
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import itertools

from sage.rings.integer_ring import ZZ
from .composition import Compositions_n, Composition
from sage.rings.integer import Integer
from sage.arith.misc import binomial


class SignedCompositions(Compositions_n):
    """
    The class of signed compositions of `n`.

    EXAMPLES::

        sage: SC3 = SignedCompositions(3); SC3
        Signed compositions of 3
        sage: SC3.cardinality()
        18
        sage: len(SC3.list())
        18
        sage: SC3.first()
        [1, 1, 1]
        sage: SC3.last()
        [-3]
        sage: SC3.random_element() # random
        [1, -1, 1]
        sage: SC3.list()
        [[1, 1, 1],
         [1, 1, -1],
         [1, -1, 1],
         [1, -1, -1],
         [-1, 1, 1],
         [-1, 1, -1],
         [-1, -1, 1],
         [-1, -1, -1],
         [1, 2],
         [1, -2],
         [-1, 2],
         [-1, -2],
         [2, 1],
         [2, -1],
         [-2, 1],
         [-2, -1],
         [3],
         [-3]]

    TESTS::

        sage: SC = SignedCompositions(3)
        sage: TestSuite(SC).run()
    """

    def __repr__(self):
        """
        TESTS::

            sage: SignedCompositions(3)
            Signed compositions of 3
        """
        return "Signed compositions of %s" % self.n

    def __contains__(self, x):
        """
        TESTS::

            sage: [] in SignedCompositions(0)
            True
            sage: [0] in SignedCompositions(0)
            False
            sage: [2,1,3] in SignedCompositions(6)
            True
            sage: [-2, 1, -3] in SignedCompositions(6)
            True
        """
        if isinstance(x, list):
            for z in x:
                if (not isinstance(z, (int, Integer))) and z not in ZZ:
                    return False
                if z == 0:
                    return False
        elif not isinstance(x, Composition):
            return False
        return sum(abs(i) for i in x) == self.n

    def cardinality(self):
        r"""
        Return the number of elements in ``self``.

        The number of signed compositions of `n` is equal to

        .. MATH::

            \sum_{i=1}^{n+1} \binom{n-1}{i-1} 2^i

        EXAMPLES::

            sage: SC4 = SignedCompositions(4)
            sage: SC4.cardinality() == len(SC4.list())
            True
            sage: SignedCompositions(3).cardinality()
            18
        """
        return ZZ.sum(binomial(self.n - 1, i - 1) * 2**i
                      for i in range(1, self.n + 1))

    def __iter__(self):
        """
        TESTS::

            sage: SignedCompositions(0).list()   #indirect doctest
            [[]]
            sage: SignedCompositions(1).list()   #indirect doctest
            [[1], [-1]]
            sage: SignedCompositions(2).list()   #indirect doctest
            [[1, 1], [1, -1], [-1, 1], [-1, -1], [2], [-2]]
        """
        for comp in Compositions_n.__iter__(self):
            l = len(comp)
            for sign in itertools.product([1, -1], repeat=l):
                yield [sign[i] * comp[i] for i in range(l)]


from sage.misc.persist import register_unpickle_override
register_unpickle_override('sage.combinat.composition_signed', 'SignedCompositions_n', SignedCompositions)



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Fully commutative stable Grothendieck crystal

AUTHORS:

- Jianping Pan (2020-08-31): initial version

- Wencin Poh (2020-08-31): initial version

- Anne Schilling (2020-08-31): initial version
"""

# ****************************************************************************
#       Copyright (C) 2020 Jianping Pan <jppan at math dot ucdavis dot edu>
#                          Wencin Poh <wpoh at ucdavis dot edu>
#                          Anne Schilling <anne at math dot ucdavis dot edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.element import Element
from sage.structure.parent import Parent
from sage.structure.richcmp import richcmp_by_eq_and_lt
from sage.structure.unique_representation import UniqueRepresentation
from sage.categories.classical_crystals import ClassicalCrystals
from sage.categories.enumerated_sets import EnumeratedSets
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.combinat.root_system.cartan_type import CartanType
from sage.combinat import permutation
from sage.rings.integer import Integer
from sage.misc.inherit_comparison import InheritComparisonClasscallMetaclass
from sage.misc.lazy_attribute import lazy_attribute
from sage.misc.lazy_import import lazy_import

lazy_import('sage.groups.perm_gps.permgroup_named', 'SymmetricGroup')


class DecreasingHeckeFactorization(Element, metaclass=InheritComparisonClasscallMetaclass):
    """
    Class of decreasing factorizations in the 0-Hecke monoid.

    INPUT:

    - ``t`` -- decreasing factorization inputted as list of lists

    - ``max_value`` -- maximal value of entries

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
        sage: t = [[3, 2], [], [2, 1]]
        sage: h = DecreasingHeckeFactorization(t, 3); h
        (3, 2)()(2, 1)
        sage: h.excess
        1
        sage: h.factors
        3
        sage: h.max_value
        3
        sage: h.value
        ((3, 2), (), (2, 1))

        sage: u = [[3, 2, 1], [3], [2, 1]]
        sage: h = DecreasingHeckeFactorization(u); h
        (3, 2, 1)(3)(2, 1)
        sage: h.weight()
        (2, 1, 3)
        sage: h.parent()
        Decreasing Hecke factorizations with 3 factors associated to [2, 1, 3, 2, 1] with excess 1
    """
    @staticmethod
    def __classcall_private__(self, t, max_value=None, parent=None):
        """
        Assign the correct parent for ``t`` and call ``t`` as an element of that parent.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: h1 = DecreasingHeckeFactorization([[3, 1], [], [3, 2]])
            sage: h1.parent()
            Fully commutative stable Grothendieck crystal of type A_2 associated to [1, 3, 2] with excess 1
            sage: h2 = DecreasingHeckeFactorization(h1)
            sage: h1 == h2
            True

            sage: h1 = DecreasingHeckeFactorization([[3, 1], [2, 1], [2, 1]])
            sage: F = h1.parent(); F
            Decreasing Hecke factorizations with 3 factors associated to [1, 3, 2, 1] with excess 2
            sage: h2 = F(h1)
            sage: h1 == h2
            True

        TESTS::

            sage: DecreasingHeckeFactorization([[]])
            ()
        """
        _check_decreasing_hecke_factorization(t)
        if isinstance(t, DecreasingHeckeFactorization):
            u = t.value
            if parent is None:
                parent = t.parent()
        else:
            u = t
            if parent is None:
                if max_value is None:
                    letters = [x for factor in t for x in factor]
                    max_value = max(letters) if letters else 1
                from sage.monoids.hecke_monoid import HeckeMonoid
                S = SymmetricGroup(max_value+1)
                H = HeckeMonoid(S)
                word = H.from_reduced_word(x for factor in t for x in factor).reduced_word()
                factors = len(t)
                excess = sum(len(l) for l in t) - len(word)

                p = permutation.from_reduced_word(word)
                if p.has_pattern([3,2,1]):
                    word = S.from_reduced_word(word)
                    parent = DecreasingHeckeFactorizations(word, factors, excess)
                else:
                    word = S.from_reduced_word(word)
                    parent = FullyCommutativeStableGrothendieckCrystal(word, factors, excess)
        return parent.element_class(parent, u)

    def __init__(self, parent, t):
        """
        Initialize a decreasing factorization for ``self`` given the relevant data.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t = [[2, 1], [2], [], [1]]
            sage: h1 = DecreasingHeckeFactorization(t); h1
            (2, 1)(2)()(1)
            sage: h1.excess
            1
            sage: h2 = DecreasingHeckeFactorization(t, 2)
            sage: h2.value
            ((2, 1), (2,), (), (1,))
            sage: h1 == h2
            True

            sage: t = [[2, 1], [2], [], [3, 1]]
            sage: h = DecreasingHeckeFactorization(t, 5)
            sage: h.max_value
            5
            sage: h.factors
            4
            sage: h.w
            (1, 2, 1, 3)

        TESTS::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t1 = [[], [3, 1], [], [2, 1], [1]]
            sage: t2 = [[], [3, 1], [], [2, 1], [2]]
            sage: h1 = DecreasingHeckeFactorization(t1, 3)
            sage: h2 = DecreasingHeckeFactorization(t2, 3)
            sage: h3 = DecreasingHeckeFactorization(t1)
            sage: h1 == h2, h1 == h3
            (False, True)
            sage: h1 != h2, h1 != h3
            (True, False)
        """
        Element.__init__(self, parent)
        self.factors = parent.factors
        self.max_value = parent.max_value
        self.w = parent.w
        self.excess = parent.excess
        self.value = tuple(tuple(factors) for factors in t)

    def _repr_(self):
        """
        Return the representation of ``self``.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t = [[], [2, 1], [2], [], [2]]
            sage: h = DecreasingHeckeFactorization(t); h
            ()(2, 1)(2)()(2)
        """
        return "".join("("+repr(list(factor))[1:-1]+")" for factor in self.value)

    def __hash__(self):
        """
        Return hash of ``self``.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t = [[], [2, 1], [2], [], [2]]
            sage: h1 = DecreasingHeckeFactorization(t)
            sage: h2 = DecreasingHeckeFactorization(t, 3)
            sage: h3 = DecreasingHeckeFactorization(t, 2)
            sage: hash(h1) == hash(h2)
            False
            sage: hash(h1) == hash(h3)
            True
        """
        return hash((self.max_value, self.value))

    _richcmp_ = richcmp_by_eq_and_lt("__eq__", "__lt__")

    def __eq__(self, other):
        """
        Return ``True`` if ``self`` equals ``other`` and ``False`` otherwise.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t = [[], [2, 1], [2], [], [2]]
            sage: h1 = DecreasingHeckeFactorization(t)
            sage: h2 = DecreasingHeckeFactorization(t, 3)
            sage: h1 == h2
            True
        """
        return isinstance(self, type(other)) and self.value == other.value

    def __lt__(self, other):
        """
        Return ``True`` if ``self`` comes before ``other`` and ``False``
        otherwise.

        We say that `h_1` comes before `h_2` if either weight of `h_1 <` weight of `h_2`
        lexicographically, or if both weights of `h_1` and `h_2` are equal,
        but `h_1 < h_2` lexicographically.
        This ordering is mainly used for sorting or comparison.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t1 = [[], [2, 1], [], [2, 1], [1]]
            sage: t2 = [[], [2, 1], [], [2, 1], [2]]
            sage: t3 = [[], [2, 1], [2], [1], [1]]
            sage: h1 = DecreasingHeckeFactorization(t1)
            sage: h2 = DecreasingHeckeFactorization(t2)
            sage: h3 = DecreasingHeckeFactorization(t3)
            sage: h1 < h2
            True
            sage: h1 < h3
            False
        """
        return (self.weight(), self.value) < (other.weight(), other.value)

    def _latex_(self):
        r"""
        Return LaTeX code for ``self``.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t = [[2], [2, 1], [], [4, 3, 1]]
            sage: h = DecreasingHeckeFactorization(t, 6)
            sage: latex(h)
            \left(2\right)\left(2, 1\right)\left(\;\right)\left(4, 3, 1\right)
        """
        s = ""
        for factor in self.value:
            if factor:
                s += r"\left("+repr(list(factor))[1:-1]+r"\right)"
            else:
                s += r"\left(\;\right)"
        return s

    def weight(self):
        """
        Return the weight of ``self``.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t = [[2], [2, 1], [], [4, 3, 1]]
            sage: h = DecreasingHeckeFactorization(t, 6)
            sage: h.weight()
            (3, 0, 2, 1)
        """
        return tuple([len(l) for l in reversed(self.value)])

    def to_word(self):
        """
        Return the word associated to ``self`` in the 0-Hecke monoid.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t = [[2], [], [2, 1], [4, 3, 1]]
            sage: h = DecreasingHeckeFactorization(t)
            sage: h.to_word()
            [2, 2, 1, 4, 3, 1]
        """
        return [j for factors in self.value for j in factors]

    def to_increasing_hecke_biword(self):
        """
        Return the associated increasing Hecke biword of ``self``.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
            sage: t = [[2], [], [2, 1],[4, 3, 1]]
            sage: h = DecreasingHeckeFactorization(t, 4)
            sage: h.to_increasing_hecke_biword()
            [[1, 1, 1, 2, 2, 4], [1, 3, 4, 1, 2, 2]]
        """
        L = [[],[]]
        for j in range(len(self.value)):
            L[1] += list(self.value[-j-1][::-1])
            L[0] += [j+1]*len(self.value[-j-1])
        return L


class DecreasingHeckeFactorizations(UniqueRepresentation, Parent):
    """
    Set of decreasing factorizations in the 0-Hecke monoid.

    INPUT:

    - ``w`` -- an element in the symmetric group

    - ``factors`` -- the number of factors in the factorization

    - ``excess`` -- the total number of letters in the factorization minus the length of a reduced word for ``w``

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorizations
        sage: S = SymmetricGroup(3+1)
        sage: w = S.from_reduced_word([1, 3, 2, 1])
        sage: F = DecreasingHeckeFactorizations(w, 3, 3); F
        Decreasing Hecke factorizations with 3 factors associated to [1, 3, 2, 1] with excess 3
        sage: F.list()
        [(3, 1)(3, 1)(3, 2, 1), (3, 1)(3, 2, 1)(2, 1), (3, 2, 1)(2, 1)(2, 1)]
    """
    @staticmethod
    def __classcall_private__(cls, w, factors, excess):
        """
        Classcall to mend the input.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorizations
            sage: S = SymmetricGroup(3+1)
            sage: w = S.from_reduced_word([1, 2 ,1])
            sage: F = DecreasingHeckeFactorizations(w, 4, 1); F
            Decreasing Hecke factorizations with 4 factors associated to [1, 2, 1] with excess 1

            sage: from sage.monoids.hecke_monoid import HeckeMonoid
            sage: H = HeckeMonoid(SymmetricGroup(3+1))
            sage: w = H.from_reduced_word([1, 2 ,1])
            sage: G = DecreasingHeckeFactorizations(w, 4, 1); G
            Decreasing Hecke factorizations with 4 factors associated to [1, 2, 1] with excess 1
            sage: F is G
            True
        """
        from sage.monoids.hecke_monoid import HeckeMonoid
        if isinstance(w.parent(), SymmetricGroup):
            H = HeckeMonoid(w.parent())
            w = H.from_reduced_word(w.reduced_word())
        if (not w.reduced_word()) and excess != 0:
            raise ValueError("excess must be 0 for the empty word")
        return super().__classcall__(cls, w, factors, excess)

    def __init__(self, w, factors, excess):
        """
        Initialize a set for ``self`` given reduced word ``w`` in the symmetric group,
        number of factors ``factors`` and``excess`` extra letters.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorizations
            sage: S = SymmetricGroup(3+1)
            sage: w = S.from_reduced_word([2, 1, 3, 2, 1])
            sage: F = DecreasingHeckeFactorizations(w, 4, 2)
            sage: F.w
            (2, 1, 3, 2, 1)
            sage: F.factors
            4
            sage: F.excess
            2
            sage: F.H
            0-Hecke monoid of the Symmetric group of order 4! as a permutation group

        TESTS::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorizations
            sage: S = SymmetricGroup(2+1)
            sage: w = S.from_reduced_word([1, 2, 1])
            sage: F = DecreasingHeckeFactorizations(w, 3, 1)
            sage: TestSuite(F).run()
        """
        Parent.__init__(self, category=FiniteEnumeratedSets())
        self.w = tuple(w.reduced_word())
        self.factors = factors
        self.H = w.parent()
        self.max_value = len(self.H.gens())
        self.excess = excess

    def _repr_(self):
        """
        Return a representation of ``self``.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorizations
            sage: S = SymmetricGroup(3+1)
            sage: w = S.from_reduced_word([2, 1, 3, 2])
            sage: DecreasingHeckeFactorizations(w, 3, 1)
            Decreasing Hecke factorizations with 3 factors associated to [2, 1, 3, 2] with excess 1
        """
        return "Decreasing Hecke factorizations with {} factors associated to {} with excess {}".format(self.factors, list(self.w), self.excess)

    def list(self):
        """
        Return list of all elements of ``self``.

        EXAMPLES::

            sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorizations
            sage: S = SymmetricGroup(3+1)
            sage: w = S.from_reduced_word([1, 3, 2, 1])
            sage: F = DecreasingHeckeFactorizations(w, 3, 3)
            sage: F.list()
            [(3, 1)(3, 1)(3, 2, 1), (3, 1)(3, 2, 1)(2, 1), (3, 2, 1)(2, 1)(2, 1)]
        """
        return _generate_decreasing_hecke_factorizations(self.w, self.factors, self.excess, parent=self)

    # temporary workaround while an_element is overridden by Parent
    _an_element_ = EnumeratedSets.ParentMethods._an_element_

    Element = DecreasingHeckeFactorization


class FullyCommutativeStableGrothendieckCrystal(UniqueRepresentation, Parent):
    """
    The crystal on fully commutative decreasing factorizations in the 0-Hecke
    monoid, as introduced by [MPPS2020]_.

    INPUT:

    - ``w`` -- an element in the symmetric group or a (skew) shape

    - ``factors`` -- the number of factors in the factorization

    - ``excess`` -- the total number of letters in the factorization minus the
      length of a reduced word for ``w``

    - ``shape`` -- boolean (default: ``False``); indicator for input ``w``,
      ``True`` if ``w`` is entered as a (skew) shape and ``False`` otherwise

    EXAMPLES::

        sage: S = SymmetricGroup(3+1)
        sage: w = S.from_reduced_word([1, 3, 2])
        sage: B = crystals.FullyCommutativeStableGrothendieck(w, 3, 2); B
        Fully commutative stable Grothendieck crystal of type A_2 associated to [1, 3, 2] with excess 2
        sage: B.list()
        [(1)(3, 1)(3, 2),
         (3, 1)(1)(3, 2),
         (3, 1)(3, 1)(2),
         (3)(3, 1)(3, 2),
         (3, 1)(3)(3, 2),
         (3, 1)(3, 2)(2)]

    We can also access the crystal by specifying a skew shape::

        sage: crystals.FullyCommutativeStableGrothendieck([[2, 2], [1]], 4, 1, shape=True)
        Fully commutative stable Grothendieck crystal of type A_3 associated to [2, 1, 3] with excess 1

    We can compute the highest weight elements::

        sage: hw = [w for w in B if w.is_highest_weight()]
        sage: hw
        [(1)(3, 1)(3, 2), (3)(3, 1)(3, 2)]
        sage: hw[0].weight()
        (2, 2, 1)

    The crystal operators themselves move elements between adjacent factors::

        sage: b = hw[0]; b
        (1)(3, 1)(3, 2)
        sage: b.f(2)
        (3, 1)(1)(3, 2)
    """
    @staticmethod
    def __classcall_private__(cls, w, factors, excess, shape=False):
        """
        Classcall to mend the input.

        EXAMPLES::

            sage: A = crystals.FullyCommutativeStableGrothendieck([[3, 3], [2, 1]], 4, 1, shape=True); A
            Fully commutative stable Grothendieck crystal of type A_3 associated to [3, 2, 4] with excess 1
            sage: B = crystals.FullyCommutativeStableGrothendieck(SkewPartition([[3, 3], [2, 1]]), 4, 1, shape=True)
            sage: A is B
            True

            sage: C = crystals.FullyCommutativeStableGrothendieck((2, 1), 3, 2, shape=True); C
            Fully commutative stable Grothendieck crystal of type A_2 associated to [1, 3, 2] with excess 2
            sage: D = crystals.FullyCommutativeStableGrothendieck(Partition([2, 1]), 3, 2, shape=True)
            sage: C is D
            True
        """
        from sage.monoids.hecke_monoid import HeckeMonoid
        if shape:
            from sage.combinat.partition import _Partitions
            from sage.combinat.skew_partition import SkewPartition
            cond1 = isinstance(w, (tuple, list)) and len(w) == 2 and w[0] in _Partitions and w[1] in _Partitions
            cond2 = isinstance(w, SkewPartition)
            if cond1 or cond2:
                sh = SkewPartition([w[0], w[1]])
            elif w in _Partitions:
                sh = SkewPartition([w, []])
            else:
                raise ValueError("w needs to be a (skew) partition")
            word = _to_reduced_word(sh)
            max_value = max(word) if word else 1
            H = HeckeMonoid(SymmetricGroup(max_value+1))
            w = H.from_reduced_word(word)
        else:
            if isinstance(w.parent(), SymmetricGroup):
                H = HeckeMonoid(w.parent())
                w = H.from_reduced_word(w.reduced_word())
        if (not w.reduced_word()) and excess != 0:
            raise ValueError("excess must be 0 for the empty word")
        return super().__classcall__(cls, w, factors, excess)

    def __init__(self, w, factors, excess):
        """
        Initialize a crystal for ``self`` given reduced word ``w`` in the symmetric group,
        number of factors ``factors`` and``excess`` extra letters.

        EXAMPLES::

            sage: S = SymmetricGroup(3+1)
            sage: w = S.from_reduced_word([1, 3, 2])
            sage: B = crystals.FullyCommutativeStableGrothendieck(w, 3, 2)
            sage: B.w
            (1, 3, 2)
            sage: B.factors
            3
            sage: B.excess
            2
            sage: B.H
            0-Hecke monoid of the Symmetric group of order 4! as a permutation group

        The reduced word ``w`` should be fully commutative, that is, its
        associated permutation should avoid the pattern 321::

            sage: S = SymmetricGroup(3+1)
            sage: w = S.from_reduced_word([1, 2, 1])
            sage: B = crystals.FullyCommutativeStableGrothendieck(w, 4, 2)
            Traceback (most recent call last):
            ...
            ValueError: w should be fully commutative

        TESTS::

            sage: S = SymmetricGroup(3+1)
            sage: w = S.from_reduced_word([2, 3, 1])
            sage: B = crystals.FullyCommutativeStableGrothendieck(w, 4, 2)
            sage: TestSuite(B).run()
        """
        # Check if w is fully commutative
        word = w.reduced_word()
        p = permutation.from_reduced_word(word)
        if p.has_pattern([3,2,1]):
            raise ValueError("w should be fully commutative")

        Parent.__init__(self, category=ClassicalCrystals())
        self.w = tuple(word)
        self.factors = factors
        self.H = w.parent()
        self.max_value = len(self.H.gens())
        self.excess = excess
        self._cartan_type = CartanType(['A', self.factors-1])

    @lazy_attribute
    def module_generators(self):
        """
        Return generators for ``self`` as a crystal.

        EXAMPLES::

            sage: S = SymmetricGroup(3+1)
            sage: w = S.from_reduced_word([1, 3, 2])
            sage: B = crystals.FullyCommutativeStableGrothendieck(w, 3, 2)
            sage: B.module_generators
            ((1)(3, 1)(3, 2), (3)(3, 1)(3, 2))
            sage: C = crystals.FullyCommutativeStableGrothendieck(w, 4, 2)
            sage: C.module_generators
            (()(1)(3, 1)(3, 2),
             ()(3)(3, 1)(3, 2),
             (1)(1)(1)(3, 2),
             (1)(1)(3)(3, 2),
             (1)(3)(3)(3, 2))
        """
        return tuple(self(x).to_highest_weight()[0] for x in _lowest_weights(self.w, self.factors, self.excess, parent=self))

    def _repr_(self):
        """
        Return a representation of ``self``.

        EXAMPLES::

            sage: S = SymmetricGroup(3+1)
            sage: w = S.from_reduced_word([2, 1, 3, 2])
            sage: crystals.FullyCommutativeStableGrothendieck(w, 3, 1)
            Fully commutative stable Grothendieck crystal of type A_2 associated to [2, 1, 3, 2] with excess 1
        """
        return "Fully commutative stable Grothendieck crystal of type A_{} associated to {} with excess {}".format(self.factors-1, list(self.w), self.excess)

    class Element(DecreasingHeckeFactorization):
        def __init__(self, parent, t):
            """
            Create an instance ``self`` of element ``t``.

            This method takes into account the constraints on the word,
            the number of factors, and excess statistic associated to the parent class.

            EXAMPLES::

                sage: S = SymmetricGroup(3+1)
                sage: w = S.from_reduced_word([1, 3, 2])
                sage: B = crystals.FullyCommutativeStableGrothendieck(w, 3, 2)
                sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization
                sage: h = DecreasingHeckeFactorization([[3, 1], [3], [3, 2]], 4)
                sage: u = B(h); u.value
                ((3, 1), (3,), (3, 2))
                sage: v = B([[3, 1], [3], [3, 2]]); v.value
                ((3, 1), (3,), (3, 2))
            """
            u = t
            if isinstance(t, DecreasingHeckeFactorization):
                u = t.value
            _check_decreasing_hecke_factorization(t)
            _check_containment(t, parent)
            DecreasingHeckeFactorization.__init__(self, parent, u)

        def e(self, i):
            """
            Return the action of `e_i` on ``self`` using the rules described in [MPPS2020]_.

            EXAMPLES::

                sage: S = SymmetricGroup(4+1)
                sage: w = S.from_reduced_word([2, 1, 4, 3, 2])
                sage: B = crystals.FullyCommutativeStableGrothendieck(w, 4, 3)
                sage: h = B([[4, 2], [4, 2, 1], [3, 2], [2]]); h
                (4, 2)(4, 2, 1)(3, 2)(2)
                sage: h.e(1)
                (4, 2)(4, 2, 1)(3)(3, 2)
                sage: h.e(2)
                (4, 2)(2, 1)(4, 3, 2)(2)
                sage: h.e(3)
            """
            P = self.parent()
            m = P.factors
            L = list(self.value[m-i-1])
            R = list(self.value[m-i])
            b = self.bracketing(i)
            if not b[0]:
                return None
            y = b[0][-1]
            if y-1 in L and y-1 in R:
                # special case: (--x+1--)(--x+1,x--) -->> (--x+1,x--)(--x--)
                L.remove(y-1)
            else:
                L.remove(y)
            R.append(y)
            L.sort(reverse=True)
            R.sort(reverse=True)
            s = [self.value[j] for j in range(m-i-1)]+[L]+[R]+[self.value[j] for j in range(m-i+1, m)]
            return P.element_class(P, s)

        def f(self, i):
            """
            Return the action of `f_i` on ``self`` using the rules described in [MPPS2020]_.

            EXAMPLES::

                sage: S = SymmetricGroup(4+1)
                sage: w = S.from_reduced_word([3, 2, 1, 4, 3])
                sage: B = crystals.FullyCommutativeStableGrothendieck(w, 4, 3)
                sage: h = B([[3, 2], [2, 1], [4, 3], [3, 1]]); h
                (3, 2)(2, 1)(4, 3)(3, 1)
                sage: h.f(1)
                (3, 2)(2, 1)(4, 3, 1)(3)
                sage: h.f(2)
                sage: h.f(3)
                (3, 2, 1)(1)(4, 3)(3, 1)
            """
            P = self.parent()
            m = P.factors
            L = list(self.value[m-i-1])
            R = list(self.value[m-i])
            b = self.bracketing(i)
            if not b[1]:
                return None
            x = b[1][0]
            if x+1 in L and x+1 in R:
                # special case: (--x+1--)(--x+1,x--) -->> (--x+1,x--)(--x--)
                R.remove(x+1)
            else:
                R.remove(x)
            L.append(x)
            L.sort(reverse=True)
            R.sort(reverse=True)
            s = [self.value[j] for j in range(m-i-1)]+[L]+[R]+[self.value[j] for j in range(m-i+1, m)]
            return P.element_class(P, s)

        def bracketing(self, i):
            """
            Remove all bracketed letters between `i`-th and `(i+1)`-th entry.

            EXAMPLES::

                sage: S = SymmetricGroup(4+1)
                sage: w = S.from_reduced_word([3, 2, 1, 4, 3])
                sage: B = crystals.FullyCommutativeStableGrothendieck(w, 3, 2)
                sage: h = B([[3], [4, 2, 1], [4, 3, 1]])
                sage: h.bracketing(1)
                [[], []]
                sage: h.bracketing(2)
                [[], [2, 1]]
            """
            P = self.parent()
            m = P.factors
            L = list(self.value[m-i-1])
            R = list(self.value[m-i])
            right_n = list(R)
            left_n = list(L)
            left_unbracketed = []
            while left_n:
                m = max(left_n)
                left_n.remove(m)
                l = [j for j in right_n if j >= m]
                if l:
                    right_n.remove(min(l))
                else:
                    left_unbracketed += [m]
            return [list(left_unbracketed), list(right_n)]


####################
# Helper functions #
####################

def _check_decreasing_hecke_factorization(t):
    """
    Check if ``t`` is a suitable data type for a decreasing factorization in a 0-Hecke monoid.

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _check_decreasing_hecke_factorization
        sage: _check_decreasing_hecke_factorization([[3, 2], [2, 1], [4]])
        sage: _check_decreasing_hecke_factorization([[3, 2, 2], [2, 1], [4]])
        Traceback (most recent call last):
        ...
        ValueError: each nonempty factor should be a strictly decreasing sequence
        sage: _check_decreasing_hecke_factorization([[3, 'a'], [2, 1], [4]])
        Traceback (most recent call last):
        ...
        ValueError: each nonempty factor should contain integers
        sage: _check_decreasing_hecke_factorization([[3, 2], [2, 1], 4])
        Traceback (most recent call last):
        ...
        ValueError: each factor in t should be a list or tuple
    """
    if not isinstance(t, DecreasingHeckeFactorization):
        if not isinstance(t, (tuple, list)):
            raise ValueError("t should be a list or tuple")
        for factor in t:
            if not isinstance(factor, (tuple, list)):
                raise ValueError("each factor in t should be a list or tuple")
            if not all(isinstance(x,(int, Integer)) for x in factor):
                raise ValueError("each nonempty factor should contain integers")
            for i in range(len(factor)-1):
                if factor[i] <= factor[i+1]:
                    raise ValueError("each nonempty factor should be a strictly decreasing sequence")


def _check_containment(t, parent):
    """
    Check if ``t`` is an element of ``parent``.

    EXAMPLES::

        sage: S = SymmetricGroup(3+1)
        sage: w = S.from_reduced_word([1, 3, 2])
        sage: B = crystals.FullyCommutativeStableGrothendieck(w, 3, 2)

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import DecreasingHeckeFactorization, _check_containment
        sage: h1 = DecreasingHeckeFactorization([[3, 1], [3], [3, 2]], 4)
        sage: _check_containment(h1, B)

        sage: h2 = DecreasingHeckeFactorization([[3, 1], [3], [], [3, 2]])
        sage: _check_containment(h2, B)
        Traceback (most recent call last):
        ...
        ValueError: number of factors do not match

        sage: h3 = [[3, 1], [2], [3, 2]]
        sage: _check_containment(h3, B)
        Traceback (most recent call last):
        ...
        ValueError: self and parent must be specified based on equivalent words

        sage: h4 = DecreasingHeckeFactorization([[3, 1], [3, 1], [3, 2]], 3)
        sage: _check_containment(h4, B)
        Traceback (most recent call last):
        ...
        ValueError: number of excess letters do not match
    """
    if isinstance(t, DecreasingHeckeFactorization):
        factors = t.factors
        w = t.w
        excess = t.excess
    else:
        factors = len(t)
        max_value = parent.max_value
        from sage.monoids.hecke_monoid import HeckeMonoid
        H = HeckeMonoid(SymmetricGroup(max_value+1))
        w = tuple(H.from_reduced_word(x for factor in t for x in factor).reduced_word())
        excess = sum(len(l) for l in t) - len(w)

    if factors != parent.factors:
        raise ValueError("number of factors do not match")
    if w != parent.w:
            raise ValueError("self and parent must be specified based on equivalent words")
    if excess != parent.excess:
        raise ValueError("number of excess letters do not match")


def _generate_decreasing_hecke_factorizations(w, factors, ex, weight=None, parent=None):
    """
    Generate all decreasing factorizations of word ``w`` in a 0-Hecke monoid
    with fixed excess and number of factors.

    INPUT:

    - ``w`` -- a reduced word, expressed as an iterable

    - ``factors`` -- number of factors for each decreasing factorization

    - ``ex`` -- number of extra letters in each decreasing factorizations

    - ``weight`` -- (default: ``None``) if ``None``, returns all possible
      decreasing factorizations, otherwise return all those with the specified
      weight

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _generate_decreasing_hecke_factorizations
        sage: _generate_decreasing_hecke_factorizations([1, 2, 1], 3, 1)
        [()(2, 1)(2, 1),
         (2)(1)(2, 1),
         (1)(2)(2, 1),
         (1)(1)(2, 1),
         (2, 1)()(2, 1),
         (2)(2, 1)(2),
         (1)(2, 1)(2),
         (1)(2, 1)(1),
         (2, 1)(2)(2),
         (2, 1)(2)(1),
         (2, 1)(1)(2),
         (2, 1)(2, 1)()]

        sage: _generate_decreasing_hecke_factorizations([1, 2, 1], 3, 1, weight=[1, 1, 2])
        [(2, 1)(2)(2), (2, 1)(2)(1), (2, 1)(1)(2)]
    """
    if parent is None:
        max_value = max(w) if w else 1
        S = SymmetricGroup(max_value+1)
        v = S.from_reduced_word(w)
        parent = DecreasingHeckeFactorizations(v, factors, ex)

    _canonical_word = lambda w, ex: [list(w)[0]]*ex + list(w)
    wt = lambda t:[len(factor) for factor in reversed(t)]

    L = _list_equivalent_words(_canonical_word(w, ex))
    Factors = []
    for word in L:
        F = _list_all_decreasing_runs(word, factors)
        for f in F:
            t = [[word[j] for j in range(len(word)) if f[j] == i] for i in range(factors, 0, -1)]
            if weight is None or weight == wt(t):
                Factors.append(parent.element_class(parent, t))
    return sorted(Factors, reverse=True)


def _list_all_decreasing_runs(word, m):
    """
    List all possible decreasing runs into ``m`` factors for ``word`` in a
    0-Hecke monoid.

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _list_all_decreasing_runs
        sage: _list_all_decreasing_runs([2, 1, 2, 1], 3)
        [[2, 2, 1, 1], [3, 2, 1, 1], [3, 3, 1, 1], [3, 3, 2, 1], [3, 3, 2, 2]]
    """
    from sage.combinat.integer_vector import IntegerVectors
    J = _jumps(word)
    jump_vector = [1]+[int(j in J) for j in range(1, len(word))]
    I = sorted(IntegerVectors(m-1-len(J), len(word)+1), reverse=True)
    P = [[elt[i]+jump_vector[i] for i in range(len(word))] for elt in I]
    V = [[m+1-sum(elt[:i+1]) for i in range(len(elt))] for elt in P]
    return V


def _to_reduced_word(P):
    """
    Return a reduced word associated to skew partition ``P``.

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _to_reduced_word
        sage: P = SkewPartition([[2, 2], [1]])
        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _to_reduced_word
        sage: _to_reduced_word(P)
        [2, 1, 3]

        sage: P = SkewPartition([[], []])
        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _to_reduced_word
        sage: _to_reduced_word(P)
        []

        sage: P = SkewPartition([[2, 1], []])
        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _to_reduced_word
        sage: _to_reduced_word(P)
        [1, 3, 2]
    """
    cells = P.cells()
    if not cells:
        return []
    m = max(cell[0] for cell in cells) + 1
    n = max(cell[1] for cell in cells) + 1
    L = []
    for i in range(m, -1, -1):
        for j in range(n, -1, -1):
            if (i, j) in cells:
                L += [j-i+m]
    return L


def _lowest_weights(w, factors, ex, parent=None):
    """
    Generate all decreasing factorizations in the 0-Hecke monoid that correspond
    to some valid semistandard Young tableaux.

    The semistandard Young tableaux should have at most ``factors`` columns and their
    column reading words should be equivalent to ``w`` in a 0-Hecke monoid.

    INPUT:

    - ``w`` -- a fully commutative reduced word, expressed as an iterable

    - ``factors`` -- number of factors for each decreasing factorization

    - ``ex`` -- number of extra letters in each decreasing factorizations

    - ``parent`` -- (default: ``None``) parent of the decreasing
      factorizations, automatically assigned if it is None

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _lowest_weights
        sage: _lowest_weights([1, 2, 1], 3, 1)
        Traceback (most recent call last):
        ...
        ValueError: the word w should be fully commutative

        sage: _lowest_weights([2, 1, 3, 2], 4, 3)
        [(2, 1)(3, 1)(3, 1)(2), (2, 1)(3, 1)(3, 2)(2)]

        sage: _lowest_weights([2, 1, 3, 2], 5, 3)
        [(2, 1)(3, 1)(3, 1)(2)(),
         (2, 1)(3, 1)(3, 2)(2)(),
         (2, 1)(3, 1)(1)(1)(2),
         (2, 1)(3, 1)(1)(2)(2),
         (2, 1)(3, 1)(2)(2)(2),
         (2, 1)(3, 2)(2)(2)(2)]

        sage: _lowest_weights([1, 3], 3, 1)
        [(3, 1)(1)(), (3, 1)(3)(), (1)(1)(3), (1)(3)(3)]

        sage: _lowest_weights([3, 2, 1], 5, 2)
        [(3, 2, 1)(1)(1)()()]
    """
    p = permutation.from_reduced_word(w)
    if p.has_pattern([3,2,1]):
        raise ValueError("the word w should be fully commutative")
    if parent is None:
        k = max(w)
        S = SymmetricGroup(k+1)
        word = S.from_reduced_word(w)
        parent = FullyCommutativeStableGrothendieckCrystal(word, factors, ex)

    _canonical_word = lambda w, ex: [list(w)[0]]*ex + list(w)
    L = _list_equivalent_words(_canonical_word(w, ex))

    M = []
    for v in L:
        if _is_valid_column_word(v, factors):
            J = [0] + _jumps(v) + [len(v)]
            t = [v[J[i]:J[i+1]] for i in range(len(J)-1)]
            if len(J) < factors+1:
                t += [()]*(factors+1-len(J))
            M.append(parent.element_class(parent, t))
    return sorted(M)


def _jumps(w):
    """
    Detect all positions where letters weakly increase in ``w``.

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _jumps
        sage: w = [4, 1, 2, 1, 4, 3, 2, 1, 3, 2, 2]
        sage: _jumps(w)
        [2, 4, 8, 10]
    """
    return [i+1 for i in range(len(w)-1) if w[i] <= w[i+1]]


def _is_valid_column_word(w, m=None):
    """
    Determine if ``w`` is actually a valid column reading word of some
    semistandard Young tableau with at most ``m`` columns.

    If ``m`` is None, then we determine if ``w`` is a valid column reading word
    of some semistandard Young tableau.

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _is_valid_column_word
        sage: w = [3, 2, 2, 1, 1]
        sage: _is_valid_column_word(w)
        False

        sage: w = [3, 2, 1, 1, 1]
        sage: _is_valid_column_word(w,3)
        True

        sage: w = [3, 2, 1, 1, 1]
        sage: _is_valid_column_word(w,2)
        False

        sage: w = [3, 2, 1, 3, 1]
        sage: _is_valid_column_word(w,2)
        True
    """
    J = [0] + _jumps(w) + [len(w)]
    L = [w[J[i+1]-1:J[i]:-1] for i in range(len(J)-1)]
    if all(len(L[i]) >= len(L[i+1]) for i in range(len(L)-1)):
        if m is None or len(_jumps(w)) <= m-1:
            # By construction the sequences along rows of L are strictly
            # decreasing, so it remains to verify that the sequences along
            # columns of L are weakly increasing
            return all(L[i+1][j] >= L[i][j] for i in range(len(L)-1)
                       for j in range(len(L[i+1])))
    return False


def _list_equivalent_words(w):
    """
    List all words equivalent to ``w`` in a 0-Hecke monoid.

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _list_equivalent_words
        sage: _list_equivalent_words([1, 1, 2, 1])
        [(1, 1, 2, 1),
         (1, 2, 1, 1),
         (1, 2, 1, 2),
         (1, 2, 2, 1),
         (2, 1, 1, 2),
         (2, 1, 2, 1),
         (2, 1, 2, 2),
         (2, 2, 1, 2)]

        sage: _list_equivalent_words([2,1,3,1,2])
        [(2, 1, 1, 3, 2),
         (2, 1, 3, 1, 2),
         (2, 1, 3, 2, 2),
         (2, 1, 3, 3, 2),
         (2, 2, 1, 3, 2),
         (2, 2, 3, 1, 2),
         (2, 3, 1, 1, 2),
         (2, 3, 1, 2, 2),
         (2, 3, 1, 3, 2),
         (2, 3, 3, 1, 2)]
    """
    if all(isinstance(i, (int, Integer)) for i in w):
        u = w
    else:
        raise ValueError("w needs to be a tuple of integers")

    def _applicable_relations(word):
        """
        Return all positions where a relation can be applied on ``word``
        along with the type of relation.
        """
        L = []
        for i in range(len(word)-2):
            p, q, r = word[i:(i+2)+1]
            if abs(p-q) > 1:
                L += [[i,"pq=qp"]]
            elif abs(p-q) == 1:
                if p == r:  # p != q by the abs test
                    L += [[i,"pqp=qpq"]]
            elif r != p:  # We must have p == q
                L += [[i,"ppq=pqq"]]
            if q == r and r != p:
                L += [[i,"pqq=ppq"]]
        if len(word) > 1 and abs(word[-2]-word[-1]) > 1:
            L += [[len(word)-2,"pq=qp"]]
        return L

    V = set()
    queue = [tuple(u)]
    while queue:
        v = queue.pop(0)
        if tuple(v) not in V:
            V.add(tuple(v))
            L = _applicable_relations(v)
            for pair in L:
                position, move = pair
                t = _apply_relations(v, position, move)
                queue += [tuple(t)]
    return sorted(v for v in list(V))


def _apply_relations(word, position, move):
    """
    Apply a particular type of ``move`` on ``word`` at the specified
    ``position`` using a relation in a 0-Hecke monoid .

    EXAMPLES::

        sage: from sage.combinat.crystals.fully_commutative_stable_grothendieck import _apply_relations
        sage: w = [2, 1, 3, 4]
        sage: _apply_relations(w, position=1, move='pq=qp')
        [2, 3, 1, 4]

        sage: w = [1, 3, 2, 1, 2, 4]
        sage: _apply_relations(w, position=2, move='pqp=qpq')
        [1, 3, 1, 2, 1, 4]

        sage: w = [2, 3, 1, 2, 2, 3]
        sage: _apply_relations(w, position=3, move='pp=p')
        [2, 3, 1, 2, 3]

        sage: w = [2, 3, 1, 2, 3]
        sage: _apply_relations(w, position=3, move='p=pp')
        [2, 3, 1, 2, 2, 3]

        sage: w = [2, 3, 1, 2, 2, 3]
        sage: _apply_relations(w, position=2, move='pqq=ppq')
        [2, 3, 1, 1, 2, 3]

        sage: w = [2, 3, 1, 1, 2, 3]
        sage: _apply_relations(w, position=2, move='ppq=pqq')
        [2, 3, 1, 2, 2, 3]
    """
    w = list(word)
    # Type 1
    if move == "pq=qp":
        p = w[position]
        q = w[position+1]
        w[position] = q
        w[position+1] = p
    # Type 2
    elif move == "pqp=qpq":
        p = w[position]
        q = w[position+1]
        w[position] = q
        w[position+1] = p
        w[position+2] = q
    # Type 3
    elif move == "pqq=ppq":
        p = w[position]
        q = w[position+2]
        w[position+1] = p
    # Type 4
    elif move == "ppq=pqq":
        p = w[position]
        q = w[position+2]
        w[position+1] = q
    # Type 5
    elif move == "pp=p":
        p = w[position]
        w = w[:position+1] + w[position+2:]
    elif move == "p=pp":
        p = w[position]
        w = w[:position+1] + [p] + w[position+1:]
    return w



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
r"""
Blob algebras

AUTHORS:

- Travis Scrimshaw (2020-05-16): Initial version
"""

# ****************************************************************************
#       Copyright (C) 2020 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.algebras import Algebras
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.combinat.diagram_algebras import (TemperleyLiebDiagrams, diagram_latex,
                                            TL_diagram_ascii_art)
from sage.combinat.dyck_word import DyckWords
from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.subset import powerset
from sage.misc.cachefunc import cached_method
from sage.rings.integer import Integer
from sage.structure.element import Element, get_coercion_model
from sage.structure.parent import Parent
from sage.structure.richcmp import richcmp
from sage.structure.unique_representation import UniqueRepresentation


class BlobDiagram(Element):
    r"""
    A blob diagram.

    A blob diagram consists of a perfect matching of the set
    `\{1, \ldots, n\} \sqcup \{-1, \ldots, -n\}` such that the result
    is a noncrossing matching (a :class:`Temperley-Lieb diagram
    <sage.combinat.diagram_algebras.TemperleyLiebDiagram>`), divided
    into two sets of pairs: one for the pairs with blobs and one for
    those without. The blobed pairs must either be either the leftmost
    propagating strand or to the left of it and not nested.
    """

    def __init__(self, parent, marked, unmarked):
        r"""
        Initialize ``self``.

        TESTS::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: B = BD4([[1,-3]], [[2,-4], [3,4], [-1,-2]])
            sage: TestSuite(B).run()
        """
        Element.__init__(self, parent)
        self.marked = tuple(sorted([tuple(sorted(pair)) for pair in marked]))
        self.unmarked = tuple(sorted([tuple(sorted(pair)) for pair in unmarked]))

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: BD4([[1,-3]], [[2,-4], [3,4], [-1,-2]])
            ({{-3, 1}}, {{-4, 2}, {-2, -1}, {3, 4}})
        """
        return '({{{}}}, {{{}}})'.format(', '.join('{' + repr(X)[1:-1] + '}'
                                                   for X in self.marked),
                                         ', '.join('{' + repr(X)[1:-1] + '}'
                                                   for X in self.unmarked))

    def __hash__(self):
        r"""
        Return the hash of ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: B = BD4([[1,-3]], [[2,-4], [3,4], [-1,-2]])
            sage: hash(B) in [hash(D) for D in BD4]
            True
            sage: len(set([hash(D) for D in BD4])) == len(BD4)
            True
        """
        return hash((self.marked, self.unmarked))

    def _richcmp_(self, other, op):
        r"""
        Compare ``self`` to ``other`` with operation ``op``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: B = BD4([[1,-3]], [[2,-4], [3,4], [-1,-2]])
            sage: any(B == D for D in BD4)
            True
            sage: B2 = BD4([], [[1,-3], [2,-4], [3,4], [-1,-2]])
            sage: B == B2
            False
            sage: B != B2
            True
            sage: sorted(BlobDiagrams(3))
            [({}, {{-3, -2}, {-1, 1}, {2, 3}}),
             ({}, {{-3, -2}, {-1, 3}, {1, 2}}),
             ({}, {{-3, 1}, {-2, -1}, {2, 3}}),
             ({}, {{-3, 3}, {-2, -1}, {1, 2}}),
             ({}, {{-3, 3}, {-2, 2}, {-1, 1}}),
             ({{-3, 1}}, {{-2, -1}, {2, 3}}),
             ({{-3, 3}}, {{-2, -1}, {1, 2}}),
             ({{-2, -1}}, {{-3, 1}, {2, 3}}),
             ({{-2, -1}}, {{-3, 3}, {1, 2}}),
             ({{-1, 1}}, {{-3, -2}, {2, 3}}),
             ({{-1, 1}}, {{-3, 3}, {-2, 2}}),
             ({{-1, 3}}, {{-3, -2}, {1, 2}}),
             ({{1, 2}}, {{-3, -2}, {-1, 3}}),
             ({{1, 2}}, {{-3, 3}, {-2, -1}}),
             ({{-3, 1}, {-2, -1}}, {{2, 3}}),
             ({{-3, 3}, {-2, -1}}, {{1, 2}}),
             ({{-3, 3}, {1, 2}}, {{-2, -1}}),
             ({{-2, -1}, {1, 2}}, {{-3, 3}}),
             ({{-1, 3}, {1, 2}}, {{-3, -2}}),
             ({{-3, 3}, {-2, -1}, {1, 2}}, {})]
        """
        return richcmp((len(self.marked), self.marked, self.unmarked),
                       (len(other.marked), other.marked, other.unmarked),
                       op)

    def temperley_lieb_diagram(self):
        r"""
        Return the Temperley-Lieb diagram corresponding to ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: B = BD4([[1,-3]], [[2,-4], [3,4], [-1,-2]])
            sage: B.temperley_lieb_diagram()
            {{-4, 2}, {-3, 1}, {-2, -1}, {3, 4}}
        """
        return self.parent()._TL_diagrams(self.marked + self.unmarked)


class BlobDiagrams(Parent, UniqueRepresentation):
    r"""
    The set of all blob diagrams.
    """

    def __init__(self, n):
        r"""
        Initialize ``self``.

        TESTS::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: TestSuite(BD4).run()
        """
        self._n = Integer(n)
        self._TL_diagrams = TemperleyLiebDiagrams(n)
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BlobDiagrams(4)
            Blob diagrams of order 4
        """
        return f"Blob diagrams of order {self._n}"

    def cardinality(self):
        r"""
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: BD4.cardinality()
            70
        """
        return (2 * self._n).binomial(self._n)

    def order(self):
        r"""
        Return the order of ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: BD4.order()
            4
        """
        return self._n

    @cached_method
    def base_set(self):
        r"""
        Return the base set of ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: sorted(BD4.base_set())
            [-4, -3, -2, -1, 1, 2, 3, 4]
        """
        return frozenset(range(1, self._n + 1)).union(range(-self._n, 0))

    def _element_constructor_(self, marked, unmarked=None):
        r"""
        Construct an element of ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: BD4([[1,-3]], [[-1,-2], [2,3], [-4,4]])
            ({{-3, 1}}, {{-4, 4}, {-2, -1}, {2, 3}})
            sage: BD4([[(1,-3)], ([-1,-2], (2,3), [-4,4])])
            ({{-3, 1}}, {{-4, 4}, {-2, -1}, {2, 3}})
        """
        if unmarked is None:
            marked, unmarked = marked
        ret = self.element_class(self, marked, unmarked)
        if ret not in self:
            raise ValueError("not a blob diagram of order {}".format(self._n))
        return ret

    def __contains__(self, X):
        r"""
        Check if ``X`` is contained in ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD4 = BlobDiagrams(4)
            sage: BD4([[1,-3], [-1,-2]], [[2,-4], [3,4]])  # indirect doctest
            ({{-3, 1}, {-2, -1}}, {{-4, 2}, {3, 4}})
            sage: BD4([[1,4], [-1,-2], [-3,-4]], [[2,3]])  # indirect doctest
            ({{-4, -3}, {-2, -1}, {1, 4}}, {{2, 3}})

            sage: BD4([[1,-2], [-1,-3]], [[2,-4], [3,4]])  # crossing strands
            Traceback (most recent call last):
            ...
            ValueError: not a blob diagram of order 4
            sage: BD4([[1,-4], [-1,-2]], [[2,-3], [3,4]])  # crossing strands
            Traceback (most recent call last):
            ...
            ValueError: not a blob diagram of order 4
            sage: BD4([[1,-2], [-1,-3]], [[3,-4], [2,4]])  # crossing strands
            Traceback (most recent call last):
            ...
            ValueError: not a blob diagram of order 4
            sage: BD4([[1,-3], [-1,-2], [3,4]], [[2,-4]])  # trapped blob cup
            Traceback (most recent call last):
            ...
            ValueError: not a blob diagram of order 4
            sage: BD4([[-1,3], [1,2], [-3,-4]], [[-2,4]])  # trapped blob cap
            Traceback (most recent call last):
            ...
            ValueError: not a blob diagram of order 4
            sage: BD4([[1,4], [-1,-2], [-3,-4], [2,3]], [])  # nested blob cup
            Traceback (most recent call last):
            ...
            ValueError: not a blob diagram of order 4
            sage: BD4([[-1,-4], [1,2], [3,4], [-2,-3]], [])  # nested blob cap
            Traceback (most recent call last):
            ...
            ValueError: not a blob diagram of order 4
            sage: BD4([[3,-3]], [[1,-1],[2,-2],[4,-4]])  # trapped propagating line
            Traceback (most recent call last):
            ...
            ValueError: not a blob diagram of order 4
        """
        if not isinstance(X, BlobDiagram):
            return False
        # Check that it is a Temperley-Lieb diagram
        TL = X.marked + X.unmarked  # the TL diagram
        if TL not in self._TL_diagrams:
            return False
        # Check left escaping
        for x, y in X.marked:
            if x > 0:  # Must be a cup
                for P in TL:
                    if P[1] < 0:  # P is a cap
                        continue
                    if P[1] < x:
                        if P[0] < 0:  # A propagating line to the left
                            return False
                    else:  # Note that P[1] != x
                        if 0 < P[0] < x:  # A nesting line
                            return False
            elif y < 0:  # Must be a cap
                for P in TL:
                    if P[0] > 0:  # P is a cup
                        continue
                    if P[0] > y:
                        if P[1] > 0:  # A propagating line to the left
                            return False
                    else:  # Note that P[0] != y
                        if 0 > P[1] > y:  # A nesting line
                            return False
            else:  # Must be a propagating line
                if any(P[0] < 0 and P[1] > 0 and P[1] < y for P in TL):
                    return False
        return True

    def __iter__(self):
        r"""
        Iterate over ``self``.

        EXAMPLES::

            sage: from sage.combinat.blob_algebra import BlobDiagrams
            sage: BD3 = BlobDiagrams(3)
            sage: sorted(BD3)
            [({}, {{-3, -2}, {-1, 1}, {2, 3}}),
             ({}, {{-3, -2}, {-1, 3}, {1, 2}}),
             ({}, {{-3, 1}, {-2, -1}, {2, 3}}),
             ({}, {{-3, 3}, {-2, -1}, {1, 2}}),
             ({}, {{-3, 3}, {-2, 2}, {-1, 1}}),
             ({{-3, 1}}, {{-2, -1}, {2, 3}}),
             ({{-3, 3}}, {{-2, -1}, {1, 2}}),
             ({{-2, -1}}, {{-3, 1}, {2, 3}}),
             ({{-2, -1}}, {{-3, 3}, {1, 2}}),
             ({{-1, 1}}, {{-3, -2}, {2, 3}}),
             ({{-1, 1}}, {{-3, 3}, {-2, 2}}),
             ({{-1, 3}}, {{-3, -2}, {1, 2}}),
             ({{1, 2}}, {{-3, -2}, {-1, 3}}),
             ({{1, 2}}, {{-3, 3}, {-2, -1}}),
             ({{-3, 1}, {-2, -1}}, {{2, 3}}),
             ({{-3, 3}, {-2, -1}}, {{1, 2}}),
             ({{-3, 3}, {1, 2}}, {{-2, -1}}),
             ({{-2, -1}, {1, 2}}, {{-3, 3}}),
             ({{-1, 3}, {1, 2}}, {{-3, -2}}),
             ({{-3, 3}, {-2, -1}, {1, 2}}, {})]
        """
        for D in DyckWords(self._n):
            markable = set()
            unmarked = []
            unpaired = []
            # Determine the pairing and which pairings are markable
            for i,d in enumerate(D):
                if i >= self._n:
                    i = -2*self._n + i
                else:
                    i += 1
                if d == 1:
                    unpaired.append(i)
                else: # d == 0
                    m = unpaired.pop()
                    if not unpaired:
                        markable.add((m, i))
                    else:
                        unmarked.append((m, i))
            for X in powerset(markable):
                yield self.element_class(self, X, unmarked + list(markable.difference(X)))

    Element = BlobDiagram


class BlobAlgebra(CombinatorialFreeModule):
    r"""
    The blob algebra.

    The *blob algebra* (also known as the Temperley-Lieb algebra of type `B`
    in [ILZ2018]_, but is a quotient of the Temperley-Lieb algebra of type `B`
    defined in [Graham1985]_) is a diagram-type algebra introduced in
    [MS1994]_ whose basis consists of :class:`Temperley-Lieb diagrams
    <sage.combinat.diagram_algebras.TemperleyLiebDiagram>`, noncrossing
    perfect matchings, that may contain blobs on strands that can be
    deformed so that the blob touches the left side (which we can think of
    as a frozen pole).

    The form we give here has 3 parameters, the natural one from the
    :class:`Temperley-Lieb algebra <sage.combinat.diagram_algebras.TemperleyLiebAlgebra>`,
    one for the idempotent relation, and one for a loop with a blob.

    INPUT:

    - ``k`` -- the order
    - ``q1`` -- the loop parameter
    - ``q2`` -- the idempotent parameter
    - ``q3`` -- the blob loop parameter

    EXAMPLES::

        sage: R.<q,r,s> = ZZ[]
        sage: B4 = algebras.Blob(4, q, r, s)
        sage: B = sorted(B4.basis())
        sage: B[14]
        B({{-4, -3}}, {{-2, -1}, {1, 2}, {3, 4}})
        sage: B[40]
        B({{3, 4}}, {{-4, -3}, {-2, -1}, {1, 2}})
        sage: B[14] * B[40]
        q*r*s*B({}, {{-4, -3}, {-2, -1}, {1, 2}, {3, 4}})

    REFERENCES:

    - [MS1994]_
    - [ILZ2018]_
    """
    @staticmethod
    def __classcall_private__(cls, k, q1, q2, q3, base_ring=None, prefix='B'):
        r"""
        Normalize input to ensure a unique representation.

        TESTS::

            sage: R.<q,r,s> = ZZ[]
            sage: B3 = algebras.Blob(3, q, r, s)
            sage: Bp = algebras.Blob(3, q, r, s, R, prefix='B')
            sage: B3 is Bp
            True
        """
        if base_ring is None:
            base_ring = get_coercion_model().common_parent(q1, q2, q3)
        q1 = base_ring(q1)
        q2 = base_ring(q2)
        q3 = base_ring(q3)
        return super().__classcall__(cls, k, q1, q2, q3, base_ring, prefix)

    def __init__(self, k, q1, q2, q3, base_ring, prefix):
        r"""
        Initialize ``self``.

        TESTS::

            sage: R.<q,r,s> = ZZ[]
            sage: B4 = algebras.Blob(4, q, r, s)
            sage: TestSuite(B4).run()

            sage: B3 = algebras.Blob(3, q, r, s)
            sage: B = list(B3.basis())
            sage: TestSuite(B3).run(elements=B)  # long time
        """
        self._q1 = q1
        self._q2 = q2
        self._q3 = q3
        diagrams = BlobDiagrams(k)
        cat = Algebras(base_ring.category()).FiniteDimensional().WithBasis()
        CombinatorialFreeModule.__init__(self, base_ring, diagrams, category=cat,
                                         prefix=prefix, bracket=False)

    def _ascii_art_term(self, diagram):
        r"""
        Return an ascii art representation of ``diagram``.

        EXAMPLES::

            sage: R.<q,r,s> = ZZ[]
            sage: B2 = algebras.Blob(2, q, r, s)
            sage: x = B2.an_element()
            sage: ascii_art(x)  # indirect doctest
               o o      o o      o o
            2* `-` + 3* `-` + 2* `0`
               .-.      .0.      .-.
               o o      o o      o o
        """
        return TL_diagram_ascii_art(diagram.marked+diagram.unmarked, use_unicode=False,
                                    blobs=diagram.marked)

    def _unicode_art_term(self, diagram):
        r"""
        Return a unicode art representation of ``diagram``.

        EXAMPLES::

            sage: R.<q,r,s> = ZZ[]
            sage: B2 = algebras.Blob(2, q, r, s)
            sage: x = B2.an_element()
            sage: unicode_art(x)  # indirect doctest
               ⚬ ⚬      ⚬ ⚬      ⚬ ⚬
            2* ╰─╯ + 3* ╰─╯ + 2* ╰●╯
               ╭─╮      ╭●╮      ╭─╮
               ⚬ ⚬      ⚬ ⚬      ⚬ ⚬
        """
        return TL_diagram_ascii_art(diagram.marked+diagram.unmarked, use_unicode=True,
                                    blobs=diagram.marked)

    def _latex_term(self, diagram):
        r"""
        Return a latex representation of ``diagram``.

        EXAMPLES::

            sage: R.<q,r,s> = ZZ[]
            sage: B2 = algebras.Blob(2, q, r, s)
            sage: latex(B2.an_element())  # indirect doctest
            2 \begin{tikzpicture}[scale = 0.5,thick, baseline={(0,-1ex/2)}]
            \tikzstyle{vertex} = [shape = circle, minimum size = 7pt, inner sep = 1pt]
            \node[vertex] (G--2) at (1.5, -1) [shape = circle, draw] {};
            \node[vertex] (G--1) at (0.0, -1) [shape = circle, draw] {};
            \node[vertex] (G-1) at (0.0, 1) [shape = circle, draw] {};
            \node[vertex] (G-2) at (1.5, 1) [shape = circle, draw] {};
            \draw[] (G--2) .. controls +(-0.5, 0.5) and +(0.5, 0.5) .. (G--1);
            \draw[] (G-1) .. controls +(0.5, -0.5) and +(-0.5, -0.5) .. (G-2);
            \end{tikzpicture}
             + 3 \begin{tikzpicture}[scale = 0.5,thick, baseline={(0,-1ex/2)}]
            \tikzstyle{vertex} = [shape = circle, minimum size = 7pt, inner sep = 1pt]
            \node[vertex] (G--2) at (1.5, -1) [shape = circle, draw] {};
            \node[vertex] (G--1) at (0.0, -1) [shape = circle, draw] {};
            \node[vertex] (G-1) at (0.0, 1) [shape = circle, draw] {};
            \node[vertex] (G-2) at (1.5, 1) [shape = circle, draw] {};
            \draw[blue,very thick] (G--2) .. controls +(-0.5, 0.5) and +(0.5, 0.5) .. node[midway,circle,fill,scale=0.6] {} (G--1);
            \draw[] (G-1) .. controls +(0.5, -0.5) and +(-0.5, -0.5) .. (G-2);
            \end{tikzpicture}
             + 2 \begin{tikzpicture}[scale = 0.5,thick, baseline={(0,-1ex/2)}]
            \tikzstyle{vertex} = [shape = circle, minimum size = 7pt, inner sep = 1pt]
            \node[vertex] (G-1) at (0.0, 1) [shape = circle, draw] {};
            \node[vertex] (G-2) at (1.5, 1) [shape = circle, draw] {};
            \node[vertex] (G--2) at (1.5, -1) [shape = circle, draw] {};
            \node[vertex] (G--1) at (0.0, -1) [shape = circle, draw] {};
            \draw[blue,very thick] (G-1) .. controls +(0.5, -0.5) and +(-0.5, -0.5) .. node[midway,circle,fill,scale=0.6] {} (G-2);
            \draw[] (G--2) .. controls +(-0.5, 0.5) and +(0.5, 0.5) .. (G--1);
            \end{tikzpicture}
        """
        def edge_options(P):
            if P[1] < P[0]:
                P = [P[1], P[0]]
            if tuple(P) in diagram.marked:
                return 'blue,very thick'
            return ''

        def edge_additions(P):
            if P[1] < P[0]:
                P = [P[1], P[0]]
            if tuple(P) in diagram.marked:
                return 'node[midway,circle,fill,scale=0.6] {} '
            return ''
        return diagram_latex(diagram.marked+diagram.unmarked,
                             edge_options=edge_options,
                             edge_additions=edge_additions)

    def order(self):
        r"""
        Return the order of ``self``.

        The order of a partition algebra is defined as half of the number
        of nodes in the diagrams.

        EXAMPLES::

            sage: R.<q,r,s> = ZZ[]
            sage: B4 = algebras.Blob(4, q, r, s)
            sage: B4.order()
            4
        """
        return self._indices.order()

    @cached_method
    def one_basis(self):
        r"""
        Return the index of the basis element `1`.

        EXAMPLES::

            sage: R.<q,r,s> = ZZ[]
            sage: B4 = algebras.Blob(4, q, r, s)
            sage: B4.one_basis()
            ({}, {{-4, 4}, {-3, 3}, {-2, 2}, {-1, 1}})
        """
        B = self._indices
        return B.element_class(B, [], [[i, -i] for i in range(1, self.order()+1)])

    def product_on_basis(self, top, bot):
        r"""
        Return the product of the basis elements indexed by ``top``
        and ``bot``.

        EXAMPLES::

            sage: R.<q,r,s> = ZZ[]
            sage: B4 = algebras.Blob(4, q, r, s)
            sage: B = B4.basis()
            sage: BD = sorted(B.keys())
            sage: BD[14]
            ({{-4, -3}}, {{-2, -1}, {1, 2}, {3, 4}})
            sage: BD[40]
            ({{3, 4}}, {{-4, -3}, {-2, -1}, {1, 2}})
            sage: B4.product_on_basis(BD[14], BD[40])
            q*r*s*B({}, {{-4, -3}, {-2, -1}, {1, 2}, {3, 4}})
            sage: all(len((x*y).support()) == 1 for x in B for y in B)
            True
        """
        ret_lists = [[], []]
        coeff = self.base_ring().one()
        top_marked = set(top.marked)
        top_unmarked = set(top.unmarked)
        bot_marked = set(bot.marked)
        bot_unmarked = set(bot.unmarked)

        for top_set, is_unmarked in [(top_marked, 0), (top_unmarked, 1)]:
            while top_set:
                # We are starting a new strand
                cur, stop = top_set.pop()  # note that cur < stop
                unmarked = is_unmarked
                #print(top_set, unmarked, cur, stop)
                if cur > 0:  # Both are anchored to the top
                    ret_lists[unmarked].append((cur, stop))
                    continue
                anchored = bool(stop > 0)  # Possibly only stop is anchored

                # Follow the path from cur until we either reach stop or
                #   we break out of the loop because both ends are anchored
                while anchored or cur != stop:
                    #print(anchored, unmarked, cur, stop)
                    cur = -cur  # Move cur to the bottom diagram
                    for X in bot_marked:
                        if cur in X:
                            if unmarked:
                                unmarked = 0
                            else:
                                coeff *= self._q2
                            prev = cur
                            cur = X[1-X.index(prev)]
                            bot_marked.remove(X)
                            break
                    for X in bot_unmarked:
                        if cur in X:
                            prev = cur
                            cur = X[1-X.index(prev)]
                            bot_unmarked.remove(X)
                            break
                    if cur < 0:  # cur is anchored at the bottom
                        if anchored:
                            ret_lists[unmarked].append((stop, cur))
                            break
                        else:
                            anchored = True
                            stop, cur = cur, stop # stop is now anchored to the bottom
                            continue
                    cur = -cur  # bring cur back to the top diagram
                    for X in top_marked:
                        if cur in X:
                            if unmarked:
                                unmarked = 0
                            else:
                                coeff *= self._q2
                            prev = cur
                            cur = X[1-X.index(prev)]
                            top_marked.remove(X)
                            break
                    for X in top_unmarked:
                        if cur in X:
                            prev = cur
                            cur = X[1-X.index(prev)]
                            top_unmarked.remove(X)
                            break
                    if cur > 0:  # cur is anchored at the top
                        if anchored:
                            ret_lists[unmarked].append((stop, cur))
                            break
                        else:
                            anchored = True
                            stop, cur = cur, stop # stop is now anchored to the top
                if cur == stop:  # We have found a (marked) loop
                    if unmarked:
                        coeff *= self._q1
                    else:
                        coeff *= self._q3
        # Everything remaining in the bottom sets are just anchored
        #   at the bottom, (i.e., are of the form {-i, -j}).
        ret_lists[0].extend(bot_marked)
        ret_lists[1].extend(bot_unmarked)

        if coeff == 0:
            return self.zero()
        diagram = self._indices.element_class(self._indices, ret_lists[0], ret_lists[1])
        return self._from_dict({diagram: coeff}, remove_zeros=False)
